# Copyright 2017 The TensorFlow Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import threading
import time

import numpy as np
import tensorflow as tf
from google.protobuf import message
from werkzeug import wrappers

from tensorboard.util import encoder
from tensorboard.backend import http_util
from tensorboard.backend.event_processing import plugin_asset_util as pau
from tensorboard.plugins import base_plugin
from tensorboard.plugins.beholder import file_system_tools
from tensorboard.plugins.beholder import im_util
from tensorboard.plugins.beholder import shared_config
from tensorboard.util import tb_logging

logger = tb_logging.get_logger()

DEFAULT_INFO = [{"name": "Waiting for data...",}]


class BeholderPlugin(base_plugin.TBPlugin):
    """TensorBoard plugin for viewing model data as a live video during
    training."""

    plugin_name = shared_config.PLUGIN_NAME

    def __init__(self, context):
        self._lock = threading.Lock()
        self._MULTIPLEXER = context.multiplexer
        self.PLUGIN_LOGDIR = pau.PluginDirectory(
            context.logdir, shared_config.PLUGIN_NAME
        )
        self.FPS = 10
        self._config_file_lock = threading.Lock()
        self.most_recent_frame = None
        self.most_recent_info = DEFAULT_INFO

    def get_plugin_apps(self):
        return {
            "/change-config": self._serve_change_config,
            "/beholder-frame": self._serve_beholder_frame,
            "/section-info": self._serve_section_info,
            "/ping": self._serve_ping,
            "/is-active": self._serve_is_active,
        }

    def is_active(self):
        summary_filename = "{}/{}".format(
            self.PLUGIN_LOGDIR, shared_config.SUMMARY_FILENAME
        )
        info_filename = "{}/{}".format(
            self.PLUGIN_LOGDIR, shared_config.SECTION_INFO_FILENAME
        )
        return tf.io.gfile.exists(summary_filename) and tf.io.gfile.exists(
            info_filename
        )

    def frontend_metadata(self):
        # TODO(#2338): Keep this in sync with the `registerDashboard` call
        # on the frontend until that call is removed.
        return base_plugin.FrontendMetadata(
            element_name="tf-beholder-dashboard", remove_dom=True,
        )

    def is_config_writable(self):
        try:
            if not tf.io.gfile.exists(self.PLUGIN_LOGDIR):
                tf.io.gfile.makedirs(self.PLUGIN_LOGDIR)
            config_filename = "{}/{}".format(
                self.PLUGIN_LOGDIR, shared_config.CONFIG_FILENAME
            )
            with self._config_file_lock:
                file_system_tools.write_pickle(
                    file_system_tools.read_pickle(
                        config_filename, shared_config.DEFAULT_CONFIG
                    ),
                    config_filename,
                )
            return True
        except tf.errors.PermissionDeniedError as e:
            logger.warning(
                "Unable to write Beholder config, controls will be disabled: %s",
                e,
            )
            return False

    @wrappers.Request.application
    def _serve_is_active(self, request):
        is_active = self.is_active()
        # If the plugin isn't active, don't check if the configuration is writable
        # since that will leave traces on disk; instead return True (the default).
        is_config_writable = self.is_config_writable() if is_active else True
        response = {
            "is_active": is_active,
            "is_config_writable": is_config_writable,
        }
        return http_util.Respond(request, response, "application/json")

    def _fetch_current_frame(self):
        path = "{}/{}".format(
            self.PLUGIN_LOGDIR, shared_config.SUMMARY_FILENAME
        )
        with self._lock:
            try:
                frame = file_system_tools.read_tensor_summary(path).astype(
                    np.uint8
                )
                self.most_recent_frame = frame
                return frame
            except (message.DecodeError, IOError, tf.errors.NotFoundError):
                if self.most_recent_frame is None:
                    self.most_recent_frame = im_util.get_image_relative_to_script(
                        "no-data.png"
                    )
                return self.most_recent_frame

    @wrappers.Request.application
    def _serve_change_config(self, request):
        config = {}

        for key, value in request.form.items():
            try:
                config[key] = int(value)
            except ValueError:
                if value == "false":
                    config[key] = False
                elif value == "true":
                    config[key] = True
                else:
                    config[key] = value

        self.FPS = config["FPS"]

        with self._config_file_lock:
            file_system_tools.write_pickle(
                config,
                "{}/{}".format(
                    self.PLUGIN_LOGDIR, shared_config.CONFIG_FILENAME
                ),
            )
        return http_util.Respond(
            request, {"config": config}, "application/json"
        )

    @wrappers.Request.application
    def _serve_section_info(self, request):
        path = "{}/{}".format(
            self.PLUGIN_LOGDIR, shared_config.SECTION_INFO_FILENAME
        )
        with self._lock:
            default = self.most_recent_info
        info = file_system_tools.read_pickle(path, default=default)
        if info is not default:
            with self._lock:
                self.most_recent_info = info
        return http_util.Respond(request, info, "application/json")

    def _frame_generator(self):
        while True:
            last_duration = 0

            if self.FPS == 0:
                continue
            else:
                time.sleep(max(0, 1 / (self.FPS) - last_duration))

            start_time = time.time()
            array = self._fetch_current_frame()
            image_bytes = encoder.encode_png(array)

            frame_text = b"--frame\r\n"
            content_type = b"Content-Type: image/png\r\n\r\n"

            response_content = (
                frame_text + content_type + image_bytes + b"\r\n\r\n"
            )

            last_duration = time.time() - start_time
            yield response_content

    @wrappers.Request.application
    def _serve_beholder_frame(self, request):  # pylint: disable=unused-argument
        # Thanks to Miguel Grinberg for this technique:
        # https://blog.miguelgrinberg.com/post/video-streaming-with-flask
        mimetype = "multipart/x-mixed-replace; boundary=frame"
        return http_util.Respond(
            request, self._frame_generator(), mimetype, code=200
        )

    @wrappers.Request.application
    def _serve_ping(self, request):  # pylint: disable=unused-argument
        return http_util.Respond(
            request, {"status": "alive"}, "application/json"
        )



# Copyright 2019 The TensorFlow Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""Functions for dealing with metrics."""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import os

import six

from tensorboard.plugins.hparams import api_pb2


def run_tag_from_session_and_metric(session_name, metric_name):
    """Returns a (run,tag) tuple storing the evaluations of the specified
    metric.

    Args:
      session_name: str.
      metric_name: MetricName protobuffer.
    Returns: (run, tag) tuple.
    """
    assert isinstance(session_name, six.string_types)
    assert isinstance(metric_name, api_pb2.MetricName)
    # os.path.join() will append a final slash if the group is empty; it seems
    # like multiplexer.Tensors won't recognize paths that end with a '/' so
    # we normalize the result of os.path.join() to remove the final '/' in that
    # case.
    run = os.path.normpath(os.path.join(session_name, metric_name.group))
    tag = metric_name.tag
    return run, tag



# Copyright 2018 The TensorFlow Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

"""TensorBoard helper routine for TF op evaluator.

Requires TensorFlow.
"""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import threading


class PersistentOpEvaluator(object):
    """Evaluate a fixed TensorFlow graph repeatedly, safely, efficiently.

    Extend this class to create a particular kind of op evaluator, like an
    image encoder. In `initialize_graph`, create an appropriate TensorFlow
    graph with placeholder inputs. In `run`, evaluate this graph and
    return its result. This class will manage a singleton graph and
    session to preserve memory usage, and will ensure that this graph and
    session do not interfere with other concurrent sessions.

    A subclass of this class offers a threadsafe, highly parallel Python
    entry point for evaluating a particular TensorFlow graph.

    Example usage:

        class FluxCapacitanceEvaluator(PersistentOpEvaluator):
          \"\"\"Compute the flux capacitance required for a system.

          Arguments:
            x: Available power input, as a `float`, in jigawatts.

          Returns:
            A `float`, in nanofarads.
          \"\"\"

          def initialize_graph(self):
            self._placeholder = tf.placeholder(some_dtype)
            self._op = some_op(self._placeholder)

          def run(self, x):
            return self._op.eval(feed_dict: {self._placeholder: x})

        evaluate_flux_capacitance = FluxCapacitanceEvaluator()

        for x in xs:
          evaluate_flux_capacitance(x)
    """

    def __init__(self):
        super(PersistentOpEvaluator, self).__init__()
        self._session = None
        self._initialization_lock = threading.Lock()

    def _lazily_initialize(self):
        """Initialize the graph and session, if this has not yet been done."""
        # TODO(nickfelt): remove on-demand imports once dep situation is fixed.
        import tensorflow.compat.v1 as tf

        with self._initialization_lock:
            if self._session:
                return
            graph = tf.Graph()
            with graph.as_default():
                self.initialize_graph()
            # Don't reserve GPU because libpng can't run on GPU.
            config = tf.ConfigProto(device_count={"GPU": 0})
            self._session = tf.Session(graph=graph, config=config)

    def initialize_graph(self):
        """Create the TensorFlow graph needed to compute this operation.

        This should write ops to the default graph and return `None`.
        """
        raise NotImplementedError(
            'Subclasses must implement "initialize_graph".'
        )

    def run(self, *args, **kwargs):
        """Evaluate the ops with the given input.

        When this function is called, the default session will have the
        graph defined by a previous call to `initialize_graph`. This
        function should evaluate any ops necessary to compute the result
        of the query for the given *args and **kwargs, likely returning
        the result of a call to `some_op.eval(...)`.
        """
        raise NotImplementedError('Subclasses must implement "run".')

    def __call__(self, *args, **kwargs):
        self._lazily_initialize()
        with self._session.as_default():
            return self.run(*args, **kwargs)



# -*- coding: utf-8 -*-
# Generated by the protocol buffer compiler.  DO NOT EDIT!
# source: tensorboard/compat/proto/saver.proto

import sys
_b=sys.version_info[0]<3 and (lambda x:x) or (lambda x:x.encode('latin1'))
from google.protobuf import descriptor as _descriptor
from google.protobuf import message as _message
from google.protobuf import reflection as _reflection
from google.protobuf import symbol_database as _symbol_database
# @@protoc_insertion_point(imports)

_sym_db = _symbol_database.Default()




DESCRIPTOR = _descriptor.FileDescriptor(
  name='tensorboard/compat/proto/saver.proto',
  package='tensorboard',
  syntax='proto3',
  serialized_options=_b('\n\023org.tensorflow.utilB\013SaverProtosP\001ZHgithub.com/tensorflow/tensorflow/tensorflow/go/core/core_protos_go_proto\370\001\001'),
  serialized_pb=_b('\n$tensorboard/compat/proto/saver.proto\x12\x0btensorboard\"\x9f\x02\n\x08SaverDef\x12\x1c\n\x14\x66ilename_tensor_name\x18\x01 \x01(\t\x12\x18\n\x10save_tensor_name\x18\x02 \x01(\t\x12\x17\n\x0frestore_op_name\x18\x03 \x01(\t\x12\x13\n\x0bmax_to_keep\x18\x04 \x01(\x05\x12\x0f\n\x07sharded\x18\x05 \x01(\x08\x12%\n\x1dkeep_checkpoint_every_n_hours\x18\x06 \x01(\x02\x12>\n\x07version\x18\x07 \x01(\x0e\x32-.tensorboard.SaverDef.CheckpointFormatVersion\"5\n\x17\x43heckpointFormatVersion\x12\n\n\x06LEGACY\x10\x00\x12\x06\n\x02V1\x10\x01\x12\x06\n\x02V2\x10\x02\x42q\n\x13org.tensorflow.utilB\x0bSaverProtosP\x01ZHgithub.com/tensorflow/tensorflow/tensorflow/go/core/core_protos_go_proto\xf8\x01\x01\x62\x06proto3')
)



_SAVERDEF_CHECKPOINTFORMATVERSION = _descriptor.EnumDescriptor(
  name='CheckpointFormatVersion',
  full_name='tensorboard.SaverDef.CheckpointFormatVersion',
  filename=None,
  file=DESCRIPTOR,
  values=[
    _descriptor.EnumValueDescriptor(
      name='LEGACY', index=0, number=0,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='V1', index=1, number=1,
      serialized_options=None,
      type=None),
    _descriptor.EnumValueDescriptor(
      name='V2', index=2, number=2,
      serialized_options=None,
      type=None),
  ],
  containing_type=None,
  serialized_options=None,
  serialized_start=288,
  serialized_end=341,
)
_sym_db.RegisterEnumDescriptor(_SAVERDEF_CHECKPOINTFORMATVERSION)


_SAVERDEF = _descriptor.Descriptor(
  name='SaverDef',
  full_name='tensorboard.SaverDef',
  filename=None,
  file=DESCRIPTOR,
  containing_type=None,
  fields=[
    _descriptor.FieldDescriptor(
      name='filename_tensor_name', full_name='tensorboard.SaverDef.filename_tensor_name', index=0,
      number=1, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='save_tensor_name', full_name='tensorboard.SaverDef.save_tensor_name', index=1,
      number=2, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='restore_op_name', full_name='tensorboard.SaverDef.restore_op_name', index=2,
      number=3, type=9, cpp_type=9, label=1,
      has_default_value=False, default_value=_b("").decode('utf-8'),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='max_to_keep', full_name='tensorboard.SaverDef.max_to_keep', index=3,
      number=4, type=5, cpp_type=1, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='sharded', full_name='tensorboard.SaverDef.sharded', index=4,
      number=5, type=8, cpp_type=7, label=1,
      has_default_value=False, default_value=False,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='keep_checkpoint_every_n_hours', full_name='tensorboard.SaverDef.keep_checkpoint_every_n_hours', index=5,
      number=6, type=2, cpp_type=6, label=1,
      has_default_value=False, default_value=float(0),
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
    _descriptor.FieldDescriptor(
      name='version', full_name='tensorboard.SaverDef.version', index=6,
      number=7, type=14, cpp_type=8, label=1,
      has_default_value=False, default_value=0,
      message_type=None, enum_type=None, containing_type=None,
      is_extension=False, extension_scope=None,
      serialized_options=None, file=DESCRIPTOR),
  ],
  extensions=[
  ],
  nested_types=[],
  enum_types=[
    _SAVERDEF_CHECKPOINTFORMATVERSION,
  ],
  serialized_options=None,
  is_extendable=False,
  syntax='proto3',
  extension_ranges=[],
  oneofs=[
  ],
  serialized_start=54,
  serialized_end=341,
)

_SAVERDEF.fields_by_name['version'].enum_type = _SAVERDEF_CHECKPOINTFORMATVERSION
_SAVERDEF_CHECKPOINTFORMATVERSION.containing_type = _SAVERDEF
DESCRIPTOR.message_types_by_name['SaverDef'] = _SAVERDEF
_sym_db.RegisterFileDescriptor(DESCRIPTOR)

SaverDef = _reflection.GeneratedProtocolMessageType('SaverDef', (_message.Message,), {
  'DESCRIPTOR' : _SAVERDEF,
  '__module__' : 'tensorboard.compat.proto.saver_pb2'
  # @@protoc_insertion_point(class_scope:tensorboard.SaverDef)
  })
_sym_db.RegisterMessage(SaverDef)


DESCRIPTOR._options = None
# @@protoc_insertion_point(module_scope)



# Copyright 2019 The TensorFlow Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""Internal path prefix support for TensorBoard.

Using a path prefix of `/foo/bar` enables TensorBoard to serve from
`http://localhost:6006/foo/bar/` rather than `http://localhost:6006/`.
See the `--path_prefix` flag docs for more details.
"""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

from tensorboard import errors


class PathPrefixMiddleware(object):
    """WSGI middleware for path prefixes.

    All requests to this middleware must begin with the specified path
    prefix (otherwise, a 404 will be returned immediately). Requests
    will be forwarded to the underlying application with the path prefix
    stripped and appended to `SCRIPT_NAME` (see the WSGI spec, PEP 3333,
    for details).
    """

    def __init__(self, application, path_prefix):
        """Initializes this middleware.

        Args:
          application: The WSGI application to wrap (see PEP 3333).
          path_prefix: A string path prefix to be stripped from incoming
            requests. If empty, this middleware is a no-op. If non-empty,
            the path prefix must start with a slash and not end with one
            (e.g., "/tensorboard").
        """
        if path_prefix.endswith("/"):
            raise ValueError(
                "Path prefix must not end with slash: %r" % path_prefix
            )
        if path_prefix and not path_prefix.startswith("/"):
            raise ValueError(
                "Non-empty path prefix must start with slash: %r" % path_prefix
            )
        self._application = application
        self._path_prefix = path_prefix
        self._strict_prefix = self._path_prefix + "/"

    def __call__(self, environ, start_response):
        path = environ.get("PATH_INFO", "")
        if path != self._path_prefix and not path.startswith(
            self._strict_prefix
        ):
            raise errors.NotFoundError()
        environ["PATH_INFO"] = path[len(self._path_prefix) :]
        environ["SCRIPT_NAME"] = (
            environ.get("SCRIPT_NAME", "") + self._path_prefix
        )
        return self._application(environ, start_response)



# Copyright 2019 The TensorFlow Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""An implementation of DataProvider that serves tfdbg v2 data.

This implementation is:
  1. Based on reading data from a DebugEvent file set on the local filesystem.
  2. Implements only the relevant methods for the debugger v2 plugin, including
     - list_runs()
     - read_blob_sequences()
     - read_blob()

This class is a short-term hack. To be used in production, it awaits integration
with a more complete implementation of DataProvider such as
MultiplexerDataProvider.
"""

import json

from tensorboard.data import provider

from tensorboard.plugins.debugger_v2 import debug_data_multiplexer


PLUGIN_NAME = "debugger-v2"

ALERTS_BLOB_TAG_PREFIX = "alerts"
EXECUTION_DIGESTS_BLOB_TAG_PREFIX = "execution_digests"
EXECUTION_DATA_BLOB_TAG_PREFIX = "execution_data"
GRAPH_EXECUTION_DIGESTS_BLOB_TAG_PREFIX = "graphexec_digests"
GRAPH_EXECUTION_DATA_BLOB_TAG_PREFIX = "graphexec_data"
GRAPH_INFO_BLOB_TAG_PREFIX = "graph_info"
GRAPH_OP_INFO_BLOB_TAG_PREFIX = "graph_op_info"
SOURCE_FILE_LIST_BLOB_TAG = "source_file_list"
SOURCE_FILE_BLOB_TAG_PREFIX = "source_file"
STACK_FRAMES_BLOB_TAG_PREFIX = "stack_frames"


def alerts_run_tag_filter(run, begin, end, alert_type=None):
    """Create a RunTagFilter for Alerts.

    Args:
      run: tfdbg2 run name.
      begin: Beginning index of alerts.
      end: Ending index of alerts.
      alert_type: Optional alert type, used to restrict retrieval of alerts
        data to a single type of alerts.

    Returns:
      `RunTagFilter` for the run and range of Alerts.
    """
    tag = "%s_%d_%d" % (ALERTS_BLOB_TAG_PREFIX, begin, end)
    if alert_type is not None:
        tag += "_%s" % alert_type
    return provider.RunTagFilter(runs=[run], tags=[tag])


def _parse_alerts_blob_key(blob_key):
    """Parse the BLOB key for Alerts.

    Args:
      blob_key: The BLOB key to parse. By contract, it should have the format:
       - `${ALERTS_BLOB_TAG_PREFIX}_${begin}_${end}.${run_id}` when there is no
         alert type filter.
       - `${ALERTS_BLOB_TAG_PREFIX}_${begin}_${end}_${alert_filter}.${run_id}`
         when there is an alert type filter.

    Returns:
      - run ID
      - begin index
      - end index
      - alert_type: alert type string used to filter retrieved alert data.
          `None` if no filtering is used.
    """
    key_body, run = blob_key.split(".", 1)
    key_body = key_body[len(ALERTS_BLOB_TAG_PREFIX) :]
    key_items = key_body.split("_", 3)
    begin = int(key_items[1])
    end = int(key_items[2])
    alert_type = None
    if len(key_items) > 3:
        alert_type = key_items[3]
    return run, begin, end, alert_type


def execution_digest_run_tag_filter(run, begin, end):
    """Create a RunTagFilter for ExecutionDigests.

    This differs from `execution_data_run_tag_filter()` in that it is for
    the small-size digest objects for execution debug events, instead of the
    full-size data objects.

    Args:
      run: tfdbg2 run name.
      begin: Beginning index of ExecutionDigests.
      end: Ending index of ExecutionDigests.

    Returns:
      `RunTagFilter` for the run and range of ExecutionDigests.
    """
    return provider.RunTagFilter(
        runs=[run],
        tags=["%s_%d_%d" % (EXECUTION_DIGESTS_BLOB_TAG_PREFIX, begin, end)],
    )


def _parse_execution_digest_blob_key(blob_key):
    """Parse the BLOB key for ExecutionDigests.

    This differs from `_parse_execution_data_blob_key()` in that it is for
    the small-size digest objects for execution debug events, instead of the
    full-size data objects.

    Args:
      blob_key: The BLOB key to parse. By contract, it should have the format:
       `${EXECUTION_DIGESTS_BLOB_TAG_PREFIX}_${begin}_${end}.${run_id}`

    Returns:
      - run ID
      - begin index
      - end index
    """

    key_body, run = blob_key.split(".", 1)
    key_body = key_body[len(EXECUTION_DIGESTS_BLOB_TAG_PREFIX) :]
    begin = int(key_body.split("_")[1])
    end = int(key_body.split("_")[2])
    return run, begin, end


def execution_data_run_tag_filter(run, begin, end):
    """Create a RunTagFilter for Execution data objects.

    This differs from `execution_digest_run_tag_filter()` in that it is
    for the detailed data objects for execution, instead of the digests.

    Args:
      run: tfdbg2 run name.
      begin: Beginning index of Execution.
      end: Ending index of Execution.

    Returns:
      `RunTagFilter` for the run and range of ExecutionDigests.
    """
    return provider.RunTagFilter(
        runs=[run],
        tags=["%s_%d_%d" % (EXECUTION_DATA_BLOB_TAG_PREFIX, begin, end)],
    )


def _parse_execution_data_blob_key(blob_key):
    """Parse the BLOB key for Execution data objects.

    This differs from `_parse_execution_digest_blob_key()` in that it is
    for the deatiled data objects for execution, instead of the digests.

    Args:
      blob_key: The BLOB key to parse. By contract, it should have the format:
       `${EXECUTION_DATA_BLOB_TAG_PREFIX}_${begin}_${end}.${run_id}`

    Returns:
      - run ID
      - begin index
      - end index
    """
    key_body, run = blob_key.split(".", 1)
    key_body = key_body[len(EXECUTION_DATA_BLOB_TAG_PREFIX) :]
    begin = int(key_body.split("_")[1])
    end = int(key_body.split("_")[2])
    return run, begin, end


def graph_execution_digest_run_tag_filter(run, begin, end, trace_id=None):
    """Create a RunTagFilter for GraphExecutionTraceDigests.

    This differs from `graph_execution_data_run_tag_filter()` in that it is for
    the small-size digest objects for intra-graph execution debug events, instead
    of the full-size data objects.

    Args:
      run: tfdbg2 run name.
      begin: Beginning index of GraphExecutionTraceDigests.
      end: Ending index of GraphExecutionTraceDigests.

    Returns:
      `RunTagFilter` for the run and range of GraphExecutionTraceDigests.
    """
    # TODO(cais): Implement support for trace_id once joining of eager
    # execution and intra-graph execution is supported by DebugDataReader.
    if trace_id is not None:
        raise NotImplementedError(
            "trace_id support for graph_execution_digest_run_tag_filter() is "
            "not implemented yet."
        )
    return provider.RunTagFilter(
        runs=[run],
        tags=[
            "%s_%d_%d" % (GRAPH_EXECUTION_DIGESTS_BLOB_TAG_PREFIX, begin, end)
        ],
    )


def _parse_graph_execution_digest_blob_key(blob_key):
    """Parse the BLOB key for GraphExecutionTraceDigests.

    This differs from `_parse_graph_execution_data_blob_key()` in that it is for
    the small-size digest objects for intra-graph execution debug events,
    instead of the full-size data objects.

    Args:
      blob_key: The BLOB key to parse. By contract, it should have the format:
       `${GRAPH_EXECUTION_DIGESTS_BLOB_TAG_PREFIX}_${begin}_${end}.${run_id}`

    Returns:
      - run ID
      - begin index
      - end index
    """
    # TODO(cais): Support parsing trace_id when it is supported.
    key_body, run = blob_key.split(".", 1)
    key_body = key_body[len(GRAPH_EXECUTION_DIGESTS_BLOB_TAG_PREFIX) :]
    begin = int(key_body.split("_")[1])
    end = int(key_body.split("_")[2])
    return run, begin, end


def graph_execution_data_run_tag_filter(run, begin, end, trace_id=None):
    """Create a RunTagFilter for GraphExecutionTrace.

    This method differs from `graph_execution_digest_run_tag_filter()` in that
    it is for full-sized data objects for intra-graph execution events.

    Args:
      run: tfdbg2 run name.
      begin: Beginning index of GraphExecutionTrace.
      end: Ending index of GraphExecutionTrace.

    Returns:
      `RunTagFilter` for the run and range of GraphExecutionTrace.
    """
    # TODO(cais): Implement support for trace_id once joining of eager
    # execution and intra-graph execution is supported by DebugDataReader.
    if trace_id is not None:
        raise NotImplementedError(
            "trace_id support for graph_execution_data_run_tag_filter() is "
            "not implemented yet."
        )
    return provider.RunTagFilter(
        runs=[run],
        tags=["%s_%d_%d" % (GRAPH_EXECUTION_DATA_BLOB_TAG_PREFIX, begin, end)],
    )


def _parse_graph_execution_data_blob_key(blob_key):
    """Parse the BLOB key for GraphExecutionTrace.

    This method differs from `_parse_graph_execution_digest_blob_key()` in that
    it is for full-sized data objects for intra-graph execution events.

    Args:
      blob_key: The BLOB key to parse. By contract, it should have the format:
       `${GRAPH_EXECUTION_DATA_BLOB_TAG_PREFIX}_${begin}_${end}.${run_id}`

    Returns:
      - run ID
      - begin index
      - end index
    """
    # TODO(cais): Support parsing trace_id when it is supported.
    key_body, run = blob_key.split(".", 1)
    key_body = key_body[len(GRAPH_EXECUTION_DATA_BLOB_TAG_PREFIX) :]
    begin = int(key_body.split("_")[1])
    end = int(key_body.split("_")[2])
    return run, begin, end


def graph_op_info_run_tag_filter(run, graph_id, op_name):
    """Create a RunTagFilter for graph op info.

    Args:
      run: tfdbg2 run name.
      graph_id: Debugger-generated ID of the graph. This is assumed to
        be the ID of the graph that immediately encloses the op in question.
      op_name: Name of the op in question. (e.g., "Dense_1/MatMul")

    Returns:
      `RunTagFilter` for the run and range of graph op info.
    """
    if not graph_id:
        raise ValueError("graph_id must not be None or empty.")
    return provider.RunTagFilter(
        runs=[run],
        tags=["%s_%s_%s" % (GRAPH_OP_INFO_BLOB_TAG_PREFIX, graph_id, op_name)],
    )


def _parse_graph_op_info_blob_key(blob_key):
    """Parse the BLOB key for graph op info.

    Args:
      blob_key: The BLOB key to parse. By contract, it should have the format:
       `${GRAPH_OP_INFO_BLOB_TAG_PREFIX}_${graph_id}_${op_name}.${run_name}`,
      wherein
        - `graph_id` is a UUID
        - op_name conforms to the TensorFlow spec:
          `^[A-Za-z0-9.][A-Za-z0-9_.\\/>-]*$`
        - `run_name` is assumed to contain no dots (`'.'`s).

    Returns:
      - run name
      - graph_id
      - op name
    """
    # NOTE: the op_name itself may include dots, this is why we use `rindex()`
    # instead of `split()`.
    last_dot_index = blob_key.rindex(".")
    run = blob_key[last_dot_index + 1 :]
    key_body = blob_key[:last_dot_index]
    key_body = key_body[len(GRAPH_OP_INFO_BLOB_TAG_PREFIX) :]
    _, graph_id, op_name = key_body.split("_", 2)
    return run, graph_id, op_name


def graph_info_run_tag_filter(run, graph_id):
    """Create a RunTagFilter for graph info.

    Args:
      run: tfdbg2 run name.
      graph_id: Debugger-generated ID of the graph in question.

    Returns:
      `RunTagFilter` for the run and range of graph info.
    """
    if not graph_id:
        raise ValueError("graph_id must not be None or empty.")
    return provider.RunTagFilter(
        runs=[run], tags=["%s_%s" % (GRAPH_INFO_BLOB_TAG_PREFIX, graph_id)],
    )


def _parse_graph_info_blob_key(blob_key):
    """Parse the BLOB key for graph info.

    Args:
      blob_key: The BLOB key to parse. By contract, it should have the format:
       `${GRAPH_INFO_BLOB_TAG_PREFIX}_${graph_id}.${run_name}`,

    Returns:
      - run name
      - graph_id
    """
    key_body, run = blob_key.split(".")
    graph_id = key_body[len(GRAPH_INFO_BLOB_TAG_PREFIX) + 1 :]
    return run, graph_id


def source_file_list_run_tag_filter(run):
    """Create a RunTagFilter for listing source files.

    Args:
      run: tfdbg2 run name.

    Returns:
      `RunTagFilter` for listing the source files in the tfdbg2 run.
    """
    return provider.RunTagFilter(runs=[run], tags=[SOURCE_FILE_LIST_BLOB_TAG])


def _parse_source_file_list_blob_key(blob_key):
    """Parse the BLOB key for source file list.

    Args:
      blob_key: The BLOB key to parse. By contract, it should have the format:
       `${SOURCE_FILE_LIST_BLOB_TAG}.${run_id}`

    Returns:
      - run ID
    """
    return blob_key[blob_key.index(".") + 1 :]


def source_file_run_tag_filter(run, index):
    """Create a RunTagFilter for listing source files.

    Args:
      run: tfdbg2 run name.
      index: The index for the source file of which the content is to be
        accessed.

    Returns:
      `RunTagFilter` for accessing the content of the source file.
    """
    return provider.RunTagFilter(
        runs=[run], tags=["%s_%d" % (SOURCE_FILE_BLOB_TAG_PREFIX, index)],
    )


def _parse_source_file_blob_key(blob_key):
    """Parse the BLOB key for accessing the content of a source file.

    Args:
      blob_key: The BLOB key to parse. By contract, it should have the format:
       `${SOURCE_FILE_BLOB_TAG_PREFIX}_${index}.${run_id}`

    Returns:
      - run ID, as a str.
      - File index, as an int.
    """
    key_body, run = blob_key.split(".", 1)
    index = int(key_body[len(SOURCE_FILE_BLOB_TAG_PREFIX) + 1 :])
    return run, index


def stack_frames_run_tag_filter(run, stack_frame_ids):
    """Create a RunTagFilter for querying stack frames.

    Args:
      run: tfdbg2 run name.
      stack_frame_ids: The stack_frame_ids being requested.

    Returns:
      `RunTagFilter` for accessing the content of the source file.
    """
    return provider.RunTagFilter(
        runs=[run],
        # The stack-frame IDS are UUIDs, which do not contain underscores.
        # Hence it's safe to concatenate them with underscores.
        tags=[STACK_FRAMES_BLOB_TAG_PREFIX + "_" + "_".join(stack_frame_ids)],
    )


def _parse_stack_frames_blob_key(blob_key):
    """Parse the BLOB key for source file list.

    Args:
      blob_key: The BLOB key to parse. By contract, it should have the format:
       `${STACK_FRAMES_BLOB_TAG_PREFIX}_` +
       `${stack_frame_id_0}_..._${stack_frame_id_N}.${run_id}`

    Returns:
      - run ID
      - The stack frame IDs as a tuple of strings.
    """
    key_body, run = blob_key.split(".", 1)
    key_body = key_body[len(STACK_FRAMES_BLOB_TAG_PREFIX) + 1 :]
    stack_frame_ids = key_body.split("_")
    return run, stack_frame_ids


class LocalDebuggerV2DataProvider(provider.DataProvider):
    """A DataProvider implementation for tfdbg v2 data on local filesystem.

    In this implementation, `experiment_id` is assumed to be the path to the
    logdir that contains the DebugEvent file set.
    """

    def __init__(self, logdir):
        """Constructor of LocalDebuggerV2DataProvider.

        Args:
          logdir: Path to the directory from which the tfdbg v2 data will be
            loaded.
        """
        super(LocalDebuggerV2DataProvider, self).__init__()
        self._multiplexer = debug_data_multiplexer.DebuggerV2EventMultiplexer(
            logdir
        )

    def list_runs(self, ctx=None, *, experiment_id):
        """List runs available.

        Args:
          experiment_id: currently unused, because the backing
            DebuggerV2EventMultiplexer does not accommodate multiple experiments.

        Returns:
          Run names as a list of str.
        """
        return [
            provider.Run(
                run_id=run,  # use names as IDs
                run_name=run,
                start_time=self._get_first_event_timestamp(run),
            )
            for run in self._multiplexer.Runs()
        ]

    def _get_first_event_timestamp(self, run_name):
        try:
            return self._multiplexer.FirstEventTimestamp(run_name)
        except ValueError as e:
            return None

    def list_scalars(
        self, ctx=None, *, experiment_id, plugin_name, run_tag_filter=None
    ):
        del experiment_id, plugin_name, run_tag_filter  # Unused.
        raise TypeError("Debugger V2 DataProvider doesn't support scalars.")

    def read_scalars(
        self,
        ctx=None,
        *,
        experiment_id,
        plugin_name,
        downsample=None,
        run_tag_filter=None
    ):
        del experiment_id, plugin_name, downsample, run_tag_filter
        raise TypeError("Debugger V2 DataProvider doesn't support scalars.")

    def list_blob_sequences(
        self, ctx=None, *, experiment_id, plugin_name, run_tag_filter=None
    ):
        del experiment_id, plugin_name, run_tag_filter  # Unused currently.
        # TODO(cais): Implement this.
        raise NotImplementedError()

    def read_blob_sequences(
        self,
        ctx=None,
        *,
        experiment_id,
        plugin_name,
        downsample=None,
        run_tag_filter=None
    ):
        del experiment_id, downsample  # Unused.
        if plugin_name != PLUGIN_NAME:
            raise ValueError("Unsupported plugin_name: %s" % plugin_name)
        if run_tag_filter.runs is None:
            raise ValueError(
                "run_tag_filter.runs is expected to be specified, but is not."
            )
        if run_tag_filter.tags is None:
            raise ValueError(
                "run_tag_filter.tags is expected to be specified, but is not."
            )

        output = dict()
        existing_runs = self._multiplexer.Runs()
        for run in run_tag_filter.runs:
            if run not in existing_runs:
                continue
            output[run] = dict()
            for tag in run_tag_filter.tags:
                if tag.startswith(
                    (
                        ALERTS_BLOB_TAG_PREFIX,
                        EXECUTION_DIGESTS_BLOB_TAG_PREFIX,
                        EXECUTION_DATA_BLOB_TAG_PREFIX,
                        GRAPH_EXECUTION_DIGESTS_BLOB_TAG_PREFIX,
                        GRAPH_EXECUTION_DATA_BLOB_TAG_PREFIX,
                        GRAPH_INFO_BLOB_TAG_PREFIX,
                        GRAPH_OP_INFO_BLOB_TAG_PREFIX,
                        SOURCE_FILE_BLOB_TAG_PREFIX,
                        STACK_FRAMES_BLOB_TAG_PREFIX,
                    )
                ) or tag in (SOURCE_FILE_LIST_BLOB_TAG,):
                    output[run][tag] = [
                        provider.BlobReference(blob_key="%s.%s" % (tag, run))
                    ]
        return output

    def read_blob(self, ctx=None, *, blob_key):
        if blob_key.startswith(ALERTS_BLOB_TAG_PREFIX):
            run, begin, end, alert_type = _parse_alerts_blob_key(blob_key)
            return json.dumps(
                self._multiplexer.Alerts(
                    run, begin, end, alert_type_filter=alert_type
                )
            )
        elif blob_key.startswith(EXECUTION_DIGESTS_BLOB_TAG_PREFIX):
            run, begin, end = _parse_execution_digest_blob_key(blob_key)
            return json.dumps(
                self._multiplexer.ExecutionDigests(run, begin, end)
            )
        elif blob_key.startswith(EXECUTION_DATA_BLOB_TAG_PREFIX):
            run, begin, end = _parse_execution_data_blob_key(blob_key)
            return json.dumps(self._multiplexer.ExecutionData(run, begin, end))
        elif blob_key.startswith(GRAPH_EXECUTION_DIGESTS_BLOB_TAG_PREFIX):
            run, begin, end = _parse_graph_execution_digest_blob_key(blob_key)
            return json.dumps(
                self._multiplexer.GraphExecutionDigests(run, begin, end)
            )
        elif blob_key.startswith(GRAPH_EXECUTION_DATA_BLOB_TAG_PREFIX):
            run, begin, end = _parse_graph_execution_data_blob_key(blob_key)
            return json.dumps(
                self._multiplexer.GraphExecutionData(run, begin, end)
            )
        elif blob_key.startswith(GRAPH_INFO_BLOB_TAG_PREFIX):
            run, graph_id = _parse_graph_info_blob_key(blob_key)
            return json.dumps(self._multiplexer.GraphInfo(run, graph_id))
        elif blob_key.startswith(GRAPH_OP_INFO_BLOB_TAG_PREFIX):
            run, graph_id, op_name = _parse_graph_op_info_blob_key(blob_key)
            return json.dumps(
                self._multiplexer.GraphOpInfo(run, graph_id, op_name)
            )
        elif blob_key.startswith(SOURCE_FILE_LIST_BLOB_TAG):
            run = _parse_source_file_list_blob_key(blob_key)
            return json.dumps(self._multiplexer.SourceFileList(run))
        elif blob_key.startswith(SOURCE_FILE_BLOB_TAG_PREFIX):
            run, index = _parse_source_file_blob_key(blob_key)
            return json.dumps(self._multiplexer.SourceLines(run, index))
        elif blob_key.startswith(STACK_FRAMES_BLOB_TAG_PREFIX):
            run, stack_frame_ids = _parse_stack_frames_blob_key(blob_key)
            return json.dumps(
                self._multiplexer.StackFrames(run, stack_frame_ids)
            )
        else:
            raise ValueError("Unrecognized blob_key: %s" % blob_key)



# Copyright 2016 The TensorFlow Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""The Embedding Projector plugin."""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import collections
import functools
import imghdr
import math
import mimetypes
import os
import threading

import numpy as np
from werkzeug import wrappers

from google.protobuf import json_format
from google.protobuf import text_format

from tensorboard.backend.http_util import Respond
from tensorboard.compat import tf
from tensorboard.plugins import base_plugin
from tensorboard.plugins.projector import metadata
from tensorboard.plugins.projector.projector_config_pb2 import ProjectorConfig
from tensorboard.util import tb_logging

logger = tb_logging.get_logger()

# Number of tensors in the LRU cache.
_TENSOR_CACHE_CAPACITY = 1

# HTTP routes.
CONFIG_ROUTE = "/info"
TENSOR_ROUTE = "/tensor"
METADATA_ROUTE = "/metadata"
RUNS_ROUTE = "/runs"
BOOKMARKS_ROUTE = "/bookmarks"
SPRITE_IMAGE_ROUTE = "/sprite_image"

_IMGHDR_TO_MIMETYPE = {
    "bmp": "image/bmp",
    "gif": "image/gif",
    "jpeg": "image/jpeg",
    "png": "image/png",
}
_DEFAULT_IMAGE_MIMETYPE = "application/octet-stream"


class LRUCache(object):
    """LRU cache.

    Used for storing the last used tensor.
    """

    def __init__(self, size):
        if size < 1:
            raise ValueError("The cache size must be >=1")
        self._size = size
        self._dict = collections.OrderedDict()

    def get(self, key):
        try:
            value = self._dict.pop(key)
            self._dict[key] = value
            return value
        except KeyError:
            return None

    def set(self, key, value):
        if value is None:
            raise ValueError("value must be != None")
        try:
            self._dict.pop(key)
        except KeyError:
            if len(self._dict) >= self._size:
                self._dict.popitem(last=False)
        self._dict[key] = value


class EmbeddingMetadata(object):
    """Metadata container for an embedding.

    The metadata holds different columns with values used for
    visualization (color by, label by) in the "Embeddings" tab in
    TensorBoard.
    """

    def __init__(self, num_points):
        """Constructs a metadata for an embedding of the specified size.

        Args:
          num_points: Number of points in the embedding.
        """
        self.num_points = num_points
        self.column_names = []
        self.name_to_values = {}

    def add_column(self, column_name, column_values):
        """Adds a named column of metadata values.

        Args:
          column_name: Name of the column.
          column_values: 1D array/list/iterable holding the column values. Must be
              of length `num_points`. The i-th value corresponds to the i-th point.

        Raises:
          ValueError: If `column_values` is not 1D array, or of length `num_points`,
              or the `name` is already used.
        """
        # Sanity checks.
        if isinstance(column_values, list) and isinstance(
            column_values[0], list
        ):
            raise ValueError(
                '"column_values" must be a flat list, but we detected '
                "that its first entry is a list"
            )

        if isinstance(column_values, np.ndarray) and column_values.ndim != 1:
            raise ValueError(
                '"column_values" should be of rank 1, '
                "but is of rank %d" % column_values.ndim
            )
        if len(column_values) != self.num_points:
            raise ValueError(
                '"column_values" should be of length %d, but is of '
                "length %d" % (self.num_points, len(column_values))
            )
        if column_name in self.name_to_values:
            raise ValueError(
                'The column name "%s" is already used' % column_name
            )

        self.column_names.append(column_name)
        self.name_to_values[column_name] = column_values


def _read_tensor_tsv_file(fpath):
    with tf.io.gfile.GFile(fpath, "r") as f:
        tensor = []
        for line in f:
            line = line.rstrip("\n")
            if line:
                tensor.append(list(map(float, line.split("\t"))))
    return np.array(tensor, dtype="float32")


def _read_tensor_binary_file(fpath, shape):
    if len(shape) != 2:
        raise ValueError("Tensor must be 2D, got shape {}".format(shape))
    tensor = np.fromfile(fpath, dtype="float32")
    return tensor.reshape(shape)


def _assets_dir_to_logdir(assets_dir):
    sub_path = os.path.sep + metadata.PLUGINS_DIR + os.path.sep
    if sub_path in assets_dir:
        two_parents_up = os.pardir + os.path.sep + os.pardir
        return os.path.abspath(os.path.join(assets_dir, two_parents_up))
    return assets_dir


def _latest_checkpoints_changed(configs, run_path_pairs):
    """Returns true if the latest checkpoint has changed in any of the runs."""
    for run_name, assets_dir in run_path_pairs:
        if run_name not in configs:
            config = ProjectorConfig()
            config_fpath = os.path.join(assets_dir, metadata.PROJECTOR_FILENAME)
            if tf.io.gfile.exists(config_fpath):
                with tf.io.gfile.GFile(config_fpath, "r") as f:
                    file_content = f.read()
                text_format.Merge(file_content, config)
        else:
            config = configs[run_name]

        # See if you can find a checkpoint file in the logdir.
        logdir = _assets_dir_to_logdir(assets_dir)
        ckpt_path = _find_latest_checkpoint(logdir)
        if not ckpt_path:
            continue
        if config.model_checkpoint_path != ckpt_path:
            return True
    return False


def _parse_positive_int_param(request, param_name):
    """Parses and asserts a positive (>0) integer query parameter.

    Args:
      request: The Werkzeug Request object
      param_name: Name of the parameter.

    Returns:
      Param, or None, or -1 if parameter is not a positive integer.
    """
    param = request.args.get(param_name)
    if not param:
        return None
    try:
        param = int(param)
        if param <= 0:
            raise ValueError()
        return param
    except ValueError:
        return -1


def _rel_to_abs_asset_path(fpath, config_fpath):
    fpath = os.path.expanduser(fpath)
    if not os.path.isabs(fpath):
        return os.path.join(os.path.dirname(config_fpath), fpath)
    return fpath


def _using_tf():
    """Return true if we're not using the fake TF API stub implementation."""
    return tf.__version__ != "stub"


class ProjectorPlugin(base_plugin.TBPlugin):
    """Embedding projector."""

    plugin_name = metadata.PLUGIN_NAME

    def __init__(self, context):
        """Instantiates ProjectorPlugin via TensorBoard core.

        Args:
          context: A base_plugin.TBContext instance.
        """
        self.multiplexer = context.multiplexer
        self.logdir = context.logdir
        self._handlers = None
        self.readers = {}
        self.run_paths = None
        self._configs = {}
        self.old_num_run_paths = None
        self.config_fpaths = None
        self.tensor_cache = LRUCache(_TENSOR_CACHE_CAPACITY)

        # Whether the plugin is active (has meaningful data to process and serve).
        # Once the plugin is deemed active, we no longer re-compute the value
        # because doing so is potentially expensive.
        self._is_active = False

        # The running thread that is currently determining whether the plugin is
        # active. If such a thread exists, do not start a duplicate thread.
        self._thread_for_determining_is_active = None

        if self.multiplexer:
            self.run_paths = self.multiplexer.RunPaths()

    def get_plugin_apps(self):
        self._handlers = {
            RUNS_ROUTE: self._serve_runs,
            CONFIG_ROUTE: self._serve_config,
            TENSOR_ROUTE: self._serve_tensor,
            METADATA_ROUTE: self._serve_metadata,
            BOOKMARKS_ROUTE: self._serve_bookmarks,
            SPRITE_IMAGE_ROUTE: self._serve_sprite_image,
            "/index.js": functools.partial(
                self._serve_file,
                os.path.join("tf_projector_plugin", "index.js"),
            ),
            "/projector_binary.html": functools.partial(
                self._serve_file,
                os.path.join("tf_projector_plugin", "projector_binary.html"),
            ),
            "/projector_binary.js": functools.partial(
                self._serve_file,
                os.path.join("tf_projector_plugin", "projector_binary.js"),
            ),
        }
        return self._handlers

    def is_active(self):
        """Determines whether this plugin is active.

        This plugin is only active if any run has an embedding.

        Returns:
          Whether any run has embedding data to show in the projector.
        """
        if not self.multiplexer:
            return False

        if self._is_active:
            # We have already determined that the projector plugin should be active.
            # Do not re-compute that. We have no reason to later set this plugin to be
            # inactive.
            return True

        if self._thread_for_determining_is_active:
            # We are currently determining whether the plugin is active. Do not start
            # a separate thread.
            return self._is_active

        # The plugin is currently not active. The frontend might check again later.
        # For now, spin off a separate thread to determine whether the plugin is
        # active.
        new_thread = threading.Thread(
            target=self._determine_is_active,
            name="ProjectorPluginIsActiveThread",
        )
        self._thread_for_determining_is_active = new_thread
        new_thread.start()
        return False

    def frontend_metadata(self):
        return base_plugin.FrontendMetadata(
            es_module_path="/index.js", disable_reload=True,
        )

    def _determine_is_active(self):
        """Determines whether the plugin is active.

        This method is run in a separate thread so that the plugin can
        offer an immediate response to whether it is active and
        determine whether it should be active in a separate thread.
        """
        if self.configs:
            self._is_active = True
        self._thread_for_determining_is_active = None

    @property
    def configs(self):
        """Returns a map of run paths to `ProjectorConfig` protos."""
        run_path_pairs = list(self.run_paths.items())
        self._append_plugin_asset_directories(run_path_pairs)
        # Also accept the root logdir as a model checkpoint directory,
        # so that the projector still works when there are no runs.
        # (Case on `run` rather than `path` to avoid issues with
        # absolute/relative paths on any filesystems.)
        if not any(run == "." for (run, path) in run_path_pairs):
            run_path_pairs.append((".", self.logdir))
        if self._run_paths_changed() or _latest_checkpoints_changed(
            self._configs, run_path_pairs
        ):
            self.readers = {}
            self._configs, self.config_fpaths = self._read_latest_config_files(
                run_path_pairs
            )
            self._augment_configs_with_checkpoint_info()
        return self._configs

    def _run_paths_changed(self):
        num_run_paths = len(list(self.run_paths.keys()))
        if num_run_paths != self.old_num_run_paths:
            self.old_num_run_paths = num_run_paths
            return True
        return False

    def _augment_configs_with_checkpoint_info(self):
        for run, config in self._configs.items():
            for embedding in config.embeddings:
                # Normalize the name of the embeddings.
                if embedding.tensor_name.endswith(":0"):
                    embedding.tensor_name = embedding.tensor_name[:-2]
                # Find the size of embeddings associated with a tensors file.
                if embedding.tensor_path:
                    fpath = _rel_to_abs_asset_path(
                        embedding.tensor_path, self.config_fpaths[run]
                    )
                    tensor = self.tensor_cache.get((run, embedding.tensor_name))
                    if tensor is None:
                        try:
                            tensor = _read_tensor_tsv_file(fpath)
                        except UnicodeDecodeError:
                            tensor = _read_tensor_binary_file(
                                fpath, embedding.tensor_shape
                            )
                        self.tensor_cache.set(
                            (run, embedding.tensor_name), tensor
                        )
                    if not embedding.tensor_shape:
                        embedding.tensor_shape.extend(
                            [len(tensor), len(tensor[0])]
                        )

            reader = self._get_reader_for_run(run)
            if not reader:
                continue
            # Augment the configuration with the tensors in the checkpoint file.
            special_embedding = None
            if config.embeddings and not config.embeddings[0].tensor_name:
                special_embedding = config.embeddings[0]
                config.embeddings.remove(special_embedding)
            var_map = reader.get_variable_to_shape_map()
            for tensor_name, tensor_shape in var_map.items():
                if len(tensor_shape) != 2:
                    continue
                # Optimizer slot values are the same shape as embeddings
                # but are not embeddings.
                if ".OPTIMIZER_SLOT" in tensor_name:
                    continue
                embedding = self._get_embedding(tensor_name, config)
                if not embedding:
                    embedding = config.embeddings.add()
                    embedding.tensor_name = tensor_name
                    if special_embedding:
                        embedding.metadata_path = (
                            special_embedding.metadata_path
                        )
                        embedding.bookmarks_path = (
                            special_embedding.bookmarks_path
                        )
                if not embedding.tensor_shape:
                    embedding.tensor_shape.extend(tensor_shape)

        # Remove configs that do not have any valid (2D) tensors.
        runs_to_remove = []
        for run, config in self._configs.items():
            if not config.embeddings:
                runs_to_remove.append(run)
        for run in runs_to_remove:
            del self._configs[run]
            del self.config_fpaths[run]

    def _read_latest_config_files(self, run_path_pairs):
        """Reads and returns the projector config files in every run
        directory."""
        configs = {}
        config_fpaths = {}
        for run_name, assets_dir in run_path_pairs:
            config = ProjectorConfig()
            config_fpath = os.path.join(assets_dir, metadata.PROJECTOR_FILENAME)
            if tf.io.gfile.exists(config_fpath):
                with tf.io.gfile.GFile(config_fpath, "r") as f:
                    file_content = f.read()
                text_format.Merge(file_content, config)
            has_tensor_files = False
            for embedding in config.embeddings:
                if embedding.tensor_path:
                    if not embedding.tensor_name:
                        embedding.tensor_name = os.path.basename(
                            embedding.tensor_path
                        )
                    has_tensor_files = True
                    break

            if not config.model_checkpoint_path:
                # See if you can find a checkpoint file in the logdir.
                logdir = _assets_dir_to_logdir(assets_dir)
                ckpt_path = _find_latest_checkpoint(logdir)
                if not ckpt_path and not has_tensor_files:
                    continue
                if ckpt_path:
                    config.model_checkpoint_path = ckpt_path

            # Sanity check for the checkpoint file existing.
            if (
                config.model_checkpoint_path
                and _using_tf()
                and not tf.io.gfile.glob(config.model_checkpoint_path + "*")
            ):
                logger.warning(
                    'Checkpoint file "%s" not found',
                    config.model_checkpoint_path,
                )
                continue
            configs[run_name] = config
            config_fpaths[run_name] = config_fpath
        return configs, config_fpaths

    def _get_reader_for_run(self, run):
        if run in self.readers:
            return self.readers[run]

        config = self._configs[run]
        reader = None
        if config.model_checkpoint_path and _using_tf():
            try:
                reader = tf.train.load_checkpoint(config.model_checkpoint_path)
            except Exception:  # pylint: disable=broad-except
                logger.warning(
                    'Failed reading "%s"', config.model_checkpoint_path
                )
        self.readers[run] = reader
        return reader

    def _get_metadata_file_for_tensor(self, tensor_name, config):
        embedding_info = self._get_embedding(tensor_name, config)
        if embedding_info:
            return embedding_info.metadata_path
        return None

    def _get_bookmarks_file_for_tensor(self, tensor_name, config):
        embedding_info = self._get_embedding(tensor_name, config)
        if embedding_info:
            return embedding_info.bookmarks_path
        return None

    def _canonical_tensor_name(self, tensor_name):
        if ":" not in tensor_name:
            return tensor_name + ":0"
        else:
            return tensor_name

    def _get_embedding(self, tensor_name, config):
        if not config.embeddings:
            return None
        for info in config.embeddings:
            if self._canonical_tensor_name(
                info.tensor_name
            ) == self._canonical_tensor_name(tensor_name):
                return info
        return None

    def _append_plugin_asset_directories(self, run_path_pairs):
        for run, assets in self.multiplexer.PluginAssets(
            metadata.PLUGIN_ASSETS_NAME
        ).items():
            if metadata.PROJECTOR_FILENAME not in assets:
                continue
            assets_dir = os.path.join(
                self.run_paths[run],
                metadata.PLUGINS_DIR,
                metadata.PLUGIN_ASSETS_NAME,
            )
            assets_path_pair = (run, os.path.abspath(assets_dir))
            run_path_pairs.append(assets_path_pair)

    @wrappers.Request.application
    def _serve_file(self, file_path, request):
        """Returns a resource file."""
        res_path = os.path.join(os.path.dirname(__file__), file_path)
        with open(res_path, "rb") as read_file:
            mimetype = mimetypes.guess_type(file_path)[0]
            return Respond(request, read_file.read(), content_type=mimetype)

    @wrappers.Request.application
    def _serve_runs(self, request):
        """Returns a list of runs that have embeddings."""
        return Respond(request, list(self.configs.keys()), "application/json")

    @wrappers.Request.application
    def _serve_config(self, request):
        run = request.args.get("run")
        if run is None:
            return Respond(
                request, 'query parameter "run" is required', "text/plain", 400
            )
        if run not in self.configs:
            return Respond(
                request, 'Unknown run: "%s"' % run, "text/plain", 400
            )

        config = self.configs[run]
        return Respond(
            request, json_format.MessageToJson(config), "application/json"
        )

    @wrappers.Request.application
    def _serve_metadata(self, request):
        run = request.args.get("run")
        if run is None:
            return Respond(
                request, 'query parameter "run" is required', "text/plain", 400
            )

        name = request.args.get("name")
        if name is None:
            return Respond(
                request, 'query parameter "name" is required', "text/plain", 400
            )

        num_rows = _parse_positive_int_param(request, "num_rows")
        if num_rows == -1:
            return Respond(
                request,
                "query parameter num_rows must be integer > 0",
                "text/plain",
                400,
            )

        if run not in self.configs:
            return Respond(
                request, 'Unknown run: "%s"' % run, "text/plain", 400
            )

        config = self.configs[run]
        fpath = self._get_metadata_file_for_tensor(name, config)
        if not fpath:
            return Respond(
                request,
                'No metadata file found for tensor "%s" in the config file "%s"'
                % (name, self.config_fpaths[run]),
                "text/plain",
                400,
            )
        fpath = _rel_to_abs_asset_path(fpath, self.config_fpaths[run])
        if not tf.io.gfile.exists(fpath) or tf.io.gfile.isdir(fpath):
            return Respond(
                request,
                '"%s" not found, or is not a file' % fpath,
                "text/plain",
                400,
            )

        num_header_rows = 0
        with tf.io.gfile.GFile(fpath, "r") as f:
            lines = []
            # Stream reading the file with early break in case the file doesn't fit in
            # memory.
            for line in f:
                lines.append(line)
                if len(lines) == 1 and "\t" in lines[0]:
                    num_header_rows = 1
                if num_rows and len(lines) >= num_rows + num_header_rows:
                    break
        return Respond(request, "".join(lines), "text/plain")

    @wrappers.Request.application
    def _serve_tensor(self, request):
        run = request.args.get("run")
        if run is None:
            return Respond(
                request, 'query parameter "run" is required', "text/plain", 400
            )

        name = request.args.get("name")
        if name is None:
            return Respond(
                request, 'query parameter "name" is required', "text/plain", 400
            )

        num_rows = _parse_positive_int_param(request, "num_rows")
        if num_rows == -1:
            return Respond(
                request,
                "query parameter num_rows must be integer > 0",
                "text/plain",
                400,
            )

        if run not in self.configs:
            return Respond(
                request, 'Unknown run: "%s"' % run, "text/plain", 400
            )

        config = self.configs[run]

        tensor = self.tensor_cache.get((run, name))
        if tensor is None:
            # See if there is a tensor file in the config.
            embedding = self._get_embedding(name, config)

            if embedding and embedding.tensor_path:
                fpath = _rel_to_abs_asset_path(
                    embedding.tensor_path, self.config_fpaths[run]
                )
                if not tf.io.gfile.exists(fpath):
                    return Respond(
                        request,
                        'Tensor file "%s" does not exist' % fpath,
                        "text/plain",
                        400,
                    )
                try:
                    tensor = _read_tensor_tsv_file(fpath)
                except UnicodeDecodeError:
                    tensor = _read_tensor_binary_file(
                        fpath, embedding.tensor_shape
                    )
            else:
                reader = self._get_reader_for_run(run)
                if not reader or not reader.has_tensor(name):
                    return Respond(
                        request,
                        'Tensor "%s" not found in checkpoint dir "%s"'
                        % (name, config.model_checkpoint_path),
                        "text/plain",
                        400,
                    )
                try:
                    tensor = reader.get_tensor(name)
                except tf.errors.InvalidArgumentError as e:
                    return Respond(request, str(e), "text/plain", 400)

            self.tensor_cache.set((run, name), tensor)

        if num_rows:
            tensor = tensor[:num_rows]
        if tensor.dtype != "float32":
            tensor = tensor.astype(dtype="float32", copy=False)
        data_bytes = tensor.tobytes()
        return Respond(request, data_bytes, "application/octet-stream")

    @wrappers.Request.application
    def _serve_bookmarks(self, request):
        run = request.args.get("run")
        if not run:
            return Respond(
                request, 'query parameter "run" is required', "text/plain", 400
            )

        name = request.args.get("name")
        if name is None:
            return Respond(
                request, 'query parameter "name" is required', "text/plain", 400
            )

        if run not in self.configs:
            return Respond(
                request, 'Unknown run: "%s"' % run, "text/plain", 400
            )

        config = self.configs[run]
        fpath = self._get_bookmarks_file_for_tensor(name, config)
        if not fpath:
            return Respond(
                request,
                'No bookmarks file found for tensor "%s" in the config file "%s"'
                % (name, self.config_fpaths[run]),
                "text/plain",
                400,
            )
        fpath = _rel_to_abs_asset_path(fpath, self.config_fpaths[run])
        if not tf.io.gfile.exists(fpath) or tf.io.gfile.isdir(fpath):
            return Respond(
                request,
                '"%s" not found, or is not a file' % fpath,
                "text/plain",
                400,
            )

        bookmarks_json = None
        with tf.io.gfile.GFile(fpath, "rb") as f:
            bookmarks_json = f.read()
        return Respond(request, bookmarks_json, "application/json")

    @wrappers.Request.application
    def _serve_sprite_image(self, request):
        run = request.args.get("run")
        if not run:
            return Respond(
                request, 'query parameter "run" is required', "text/plain", 400
            )

        name = request.args.get("name")
        if name is None:
            return Respond(
                request, 'query parameter "name" is required', "text/plain", 400
            )

        if run not in self.configs:
            return Respond(
                request, 'Unknown run: "%s"' % run, "text/plain", 400
            )

        config = self.configs[run]
        embedding_info = self._get_embedding(name, config)

        if not embedding_info or not embedding_info.sprite.image_path:
            return Respond(
                request,
                'No sprite image file found for tensor "%s" in the config file "%s"'
                % (name, self.config_fpaths[run]),
                "text/plain",
                400,
            )

        fpath = os.path.expanduser(embedding_info.sprite.image_path)
        fpath = _rel_to_abs_asset_path(fpath, self.config_fpaths[run])
        if not tf.io.gfile.exists(fpath) or tf.io.gfile.isdir(fpath):
            return Respond(
                request,
                '"%s" does not exist or is directory' % fpath,
                "text/plain",
                400,
            )
        f = tf.io.gfile.GFile(fpath, "rb")
        encoded_image_string = f.read()
        f.close()
        image_type = imghdr.what(None, encoded_image_string)
        mime_type = _IMGHDR_TO_MIMETYPE.get(image_type, _DEFAULT_IMAGE_MIMETYPE)
        return Respond(request, encoded_image_string, mime_type)


def _find_latest_checkpoint(dir_path):
    if not _using_tf():
        return None
    try:
        ckpt_path = tf.train.latest_checkpoint(dir_path)
        if not ckpt_path:
            # Check the parent directory.
            ckpt_path = tf.train.latest_checkpoint(
                os.path.join(dir_path, os.pardir)
            )
        return ckpt_path
    except tf.errors.NotFoundError:
        return None



# Copyright 2017 The TensorFlow Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""The plugin for serving data from a TensorFlow debugger."""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import collections
import glob
import json
import os
import re
import sys
import threading

import tensorflow as tf
from werkzeug import wrappers

from tensorboard.backend import http_util
from tensorboard.backend.event_processing import event_file_loader
from tensorboard.plugins import base_plugin
from tensorboard.plugins.debugger import constants
from tensorboard.plugins.debugger import debugger_server_lib
from tensorboard.util import tb_logging
from tensorboard.util import tensor_util

logger = tb_logging.get_logger()

# HTTP routes.
_HEALTH_PILLS_ROUTE = "/health_pills"
_NUMERICS_ALERT_REPORT_ROUTE = "/numerics_alert_report"

# The POST key of HEALTH_PILLS_ROUTE for a JSON list of node names.
_NODE_NAMES_POST_KEY = "node_names"

# The POST key of HEALTH_PILLS_ROUTE for the run to retrieve health pills for.
_RUN_POST_KEY = "run"

# The default run to retrieve health pills for.
_DEFAULT_RUN = "."

# The POST key of HEALTH_PILLS_ROUTE for the specific step to retrieve health
# pills for.
_STEP_POST_KEY = "step"

# A glob pattern for files containing debugger-related events.
_DEBUGGER_EVENTS_GLOB_PATTERN = "events.debugger*"

# Encapsulates data for a single health pill.
HealthPillEvent = collections.namedtuple(
    "HealthPillEvent",
    [
        "wall_time",
        "step",
        "device_name",
        "output_slot",
        "node_name",
        "dtype",
        "shape",
        "value",
    ],
)


class DebuggerPlugin(base_plugin.TBPlugin):
    """TensorFlow Debugger plugin. Receives requests for debugger-related data.

    That data could include health pills, which unveil the status of
    tensor values.
    """

    # This string field is used by TensorBoard to generate the paths for routes
    # provided by this plugin. It must thus be URL-friendly. This field is also
    # used to uniquely identify this plugin throughout TensorBoard. See BasePlugin
    # for details.
    plugin_name = constants.DEBUGGER_PLUGIN_NAME

    def __init__(self, context):
        """Constructs a debugger plugin for TensorBoard.

        This plugin adds handlers for retrieving debugger-related data. The plugin
        also starts a debugger data server once the log directory is passed to the
        plugin via the call to get_plugin_apps.

        Args:
          context: A base_plugin.TBContext instance.
        """
        self._event_multiplexer = context.multiplexer
        self._logdir = context.logdir
        self._debugger_data_server = None
        self._grpc_port = None

    def listen(self, grpc_port):
        """Start listening on the given gRPC port.

        This method of an instance of DebuggerPlugin can be invoked at most once.
        This method is not thread safe.

        Args:
          grpc_port: port number to listen at.

        Raises:
          ValueError: If this instance is already listening at a gRPC port.
        """
        if self._grpc_port:
            raise ValueError(
                "This DebuggerPlugin instance is already listening at gRPC port %d"
                % self._grpc_port
            )
        self._grpc_port = grpc_port

        sys.stderr.write(
            "Creating DebuggerDataServer at port %d and logdir %s\n"
            % (self._grpc_port, self._logdir)
        )
        sys.stderr.flush()
        self._debugger_data_server = debugger_server_lib.DebuggerDataServer(
            self._grpc_port, self._logdir
        )

        threading.Thread(
            target=self._debugger_data_server.start_the_debugger_data_receiving_server
        ).start()

    def get_plugin_apps(self):
        """Obtains a mapping between routes and handlers.

        This function also starts a debugger data server on separate thread if the
        plugin has not started one yet.

        Returns:
          A mapping between routes and handlers (functions that respond to
          requests).
        """
        return {
            _HEALTH_PILLS_ROUTE: self._serve_health_pills_handler,
            _NUMERICS_ALERT_REPORT_ROUTE: self._serve_numerics_alert_report_handler,
        }

    def is_active(self):
        """Determines whether this plugin is active.

        This plugin is active if any health pills information is present for any
        run.

        Returns:
          A boolean. Whether this plugin is active.
        """
        return bool(
            self._grpc_port is not None
            and self._event_multiplexer
            and self._event_multiplexer.PluginRunToTagToContent(
                constants.DEBUGGER_PLUGIN_NAME
            )
        )

    def frontend_metadata(self):
        return base_plugin.FrontendMetadata(
            element_name="tf-debugger-dashboard"
        )

    @wrappers.Request.application
    def _serve_health_pills_handler(self, request):
        """A (wrapped) werkzeug handler for serving health pills.

        Accepts POST requests and responds with health pills. The request accepts
        several POST parameters:

          node_names: (required string) A JSON-ified list of node names for which
              the client would like to request health pills.
          run: (optional string) The run to retrieve health pills for. Defaults to
              '.'. This data is sent via POST (not GET) since URL length is limited.
          step: (optional integer): The session run step for which to
              retrieve health pills. If provided, the handler reads the health pills
              of that step from disk (which is slow) and produces a response with
              only health pills at that step. If not provided, the handler returns a
              response with health pills at all steps sampled by the event
              multiplexer (the fast path). The motivation here is that, sometimes,
              one desires to examine health pills at a specific step (to say find
              the first step that causes a model to blow up with NaNs).
              get_plugin_apps must be called before this slower feature is used
              because that method passes the logdir (directory path) to this plugin.

        This handler responds with a JSON-ified object mapping from node names to a
        list (of size 1) of health pill event objects, each of which has these
        properties.

        {
            'wall_time': float,
            'step': int,
            'node_name': string,
            'output_slot': int,
            # A list of 12 floats that summarizes the elements of the tensor.
            'value': float[],
        }

        Node names for which there are no health pills to be found are excluded from
        the mapping.

        Args:
          request: The request issued by the client for health pills.

        Returns:
          A werkzeug BaseResponse object.
        """
        if request.method != "POST":
            return http_util.Respond(
                request,
                "%s requests are forbidden by the debugger plugin."
                % request.method,
                "text/plain",
                code=405,
            )

        if _NODE_NAMES_POST_KEY not in request.form:
            return http_util.Respond(
                request,
                (
                    "The %r POST key was not found in the request for health pills."
                    % _NODE_NAMES_POST_KEY
                ),
                "text/plain",
                code=400,
            )

        jsonified_node_names = request.form[_NODE_NAMES_POST_KEY]
        try:
            node_names = json.loads(tf.compat.as_text(jsonified_node_names))
        except Exception as e:  # pylint: disable=broad-except
            # Different JSON libs raise different exceptions, so we just do a
            # catch-all here. This problem is complicated by how Tensorboard might be
            # run in many different environments, as it is open-source.
            # TODO(@caisq, @chihuahua): Create platform-dependent adapter to catch
            # specific types of exceptions, instead of the broad catching here.
            response = ("Could not decode node name JSON string %r: %s") % (
                jsonified_node_names,
                e,
            )
            return http_util.Respond(request, response, "text/plain", code=400)

        if not isinstance(node_names, list):
            response = ("%r is not a JSON list of node names:") % (
                jsonified_node_names
            )
            return http_util.Respond(request, response, "text/plain", code=400)

        run = request.form.get(_RUN_POST_KEY, _DEFAULT_RUN)
        step_string = request.form.get(_STEP_POST_KEY, None)
        if step_string is None:
            # Use all steps sampled by the event multiplexer (Relatively fast).
            mapping = self._obtain_sampled_health_pills(run, node_names)
        else:
            # Read disk to obtain the health pills for that step (Relatively slow).
            # Make sure that the directory for the run exists.
            # Determine the directory of events file to read.
            events_directory = self._logdir
            if run != _DEFAULT_RUN:
                # Use the directory for the specific run.
                events_directory = os.path.join(events_directory, run)

            step = int(step_string)
            try:
                mapping = self._obtain_health_pills_at_step(
                    events_directory, node_names, step
                )
            except IOError as error:
                response = "Error retrieving health pills for step %d: %s" % (
                    step,
                    error,
                )
                return http_util.Respond(
                    request, response, "text/plain", code=404
                )

        # Convert event_accumulator.HealthPillEvents to JSON-able dicts.
        jsonable_mapping = {}
        for node_name, events in mapping.items():
            jsonable_mapping[node_name] = [e._asdict() for e in events]
        return http_util.Respond(request, jsonable_mapping, "application/json")

    def _obtain_sampled_health_pills(self, run, node_names):
        """Obtains the health pills for a run sampled by the event multiplexer.

        This is much faster than the alternative path of reading health pills from
        disk.

        Args:
          run: The run to fetch health pills for.
          node_names: A list of node names for which to retrieve health pills.

        Returns:
          A dictionary mapping from node name to a list of
          event_accumulator.HealthPillEvents.
        """
        runs_to_tags_to_content = self._event_multiplexer.PluginRunToTagToContent(
            constants.DEBUGGER_PLUGIN_NAME
        )

        if run not in runs_to_tags_to_content:
            # The run lacks health pills.
            return {}

        # This is also a mapping between node name and plugin content because this
        # plugin tags by node name.
        tags_to_content = runs_to_tags_to_content[run]

        mapping = {}
        for node_name in node_names:
            if node_name not in tags_to_content:
                # This node lacks health pill data.
                continue

            health_pills = []
            for tensor_event in self._event_multiplexer.Tensors(run, node_name):
                json_string = tags_to_content[node_name]
                try:
                    content_object = json.loads(tf.compat.as_text(json_string))
                    device_name = content_object["device"]
                    output_slot = content_object["outputSlot"]
                    health_pills.append(
                        self._tensor_proto_to_health_pill(
                            tensor_event, node_name, device_name, output_slot
                        )
                    )
                except (KeyError, ValueError) as e:
                    logger.error(
                        "Could not determine device from JSON string " "%r: %r",
                        json_string,
                        e,
                    )

            mapping[node_name] = health_pills

        return mapping

    def _tensor_proto_to_health_pill(
        self, tensor_event, node_name, device, output_slot
    ):
        """Converts an event_accumulator.TensorEvent to a HealthPillEvent.

        Args:
          tensor_event: The event_accumulator.TensorEvent to convert.
          node_name: The name of the node (without the output slot).
          device: The device.
          output_slot: The integer output slot this health pill is relevant to.

        Returns:
          A HealthPillEvent.
        """
        return self._process_health_pill_value(
            wall_time=tensor_event.wall_time,
            step=tensor_event.step,
            device_name=device,
            output_slot=output_slot,
            node_name=node_name,
            tensor_proto=tensor_event.tensor_proto,
        )

    def _obtain_health_pills_at_step(self, events_directory, node_names, step):
        """Reads disk to obtain the health pills for a run at a specific step.

        This could be much slower than the alternative path of just returning all
        health pills sampled by the event multiplexer. It could take tens of minutes
        to complete this call for large graphs for big step values (in the
        thousands).

        Args:
          events_directory: The directory containing events for the desired run.
          node_names: A list of node names for which to retrieve health pills.
          step: The step to obtain health pills for.

        Returns:
          A dictionary mapping from node name to a list of health pill objects (see
          docs for _serve_health_pills_handler for properties of those objects).

        Raises:
          IOError: If no files with health pill events could be found.
        """
        # Obtain all files with debugger-related events.
        pattern = os.path.join(events_directory, _DEBUGGER_EVENTS_GLOB_PATTERN)
        file_paths = glob.glob(pattern)

        if not file_paths:
            raise IOError(
                "No events files found that matches the pattern %r." % pattern
            )

        # Sort by name (and thus by timestamp).
        file_paths.sort()

        mapping = collections.defaultdict(list)
        node_name_set = frozenset(node_names)

        for file_path in file_paths:
            should_stop = self._process_health_pill_event(
                node_name_set, mapping, step, file_path
            )
            if should_stop:
                break

        return mapping

    def _process_health_pill_event(
        self, node_name_set, mapping, target_step, file_path
    ):
        """Creates health pills out of data in an event.

        Creates health pills out of the event and adds them to the mapping.

        Args:
          node_name_set: A set of node names that are relevant.
          mapping: The mapping from node name to HealthPillEvents.
              This object may be destructively modified.
          target_step: The target step at which to obtain health pills.
          file_path: The path to the file with health pill events.

        Returns:
          Whether we should stop reading events because future events are no longer
          relevant.
        """
        events_loader = event_file_loader.EventFileLoader(file_path)
        for event in events_loader.Load():
            if not event.HasField("summary"):
                logger.warning(
                    "An event in a debugger events file lacks a summary."
                )
                continue

            if event.step < target_step:
                # This event is not of the relevant step. We perform this check
                # first because the majority of events will be eliminated from
                # consideration by this check.
                continue

            if event.step > target_step:
                # We have passed the relevant step. No need to read more events.
                return True

            for value in event.summary.value:
                # Obtain the device name from the metadata.
                summary_metadata = value.metadata
                plugin_data = summary_metadata.plugin_data
                if plugin_data.plugin_name == constants.DEBUGGER_PLUGIN_NAME:
                    try:
                        content = json.loads(
                            tf.compat.as_text(
                                summary_metadata.plugin_data.content
                            )
                        )
                    except ValueError as err:
                        logger.warning(
                            "Could not parse the JSON string containing data for "
                            "the debugger plugin: %r, %r",
                            content,
                            err,
                        )
                        continue
                    device_name = content["device"]
                    output_slot = content["outputSlot"]
                else:
                    logger.error(
                        "No debugger plugin data found for event with tag %s and node "
                        "name %s.",
                        value.tag,
                        value.node_name,
                    )
                    continue

                if not value.HasField("tensor"):
                    logger.warning(
                        "An event in a debugger events file lacks a tensor value."
                    )
                    continue

                match = re.match(
                    r"^(.*):(\d+):DebugNumericSummary$", value.node_name
                )
                if not match:
                    logger.warning(
                        (
                            "A event with a health pill has an invalid watch, (i.e., an "
                            "unexpected debug op): %r"
                        ),
                        value.node_name,
                    )
                    return None

                health_pill = self._process_health_pill_value(
                    wall_time=event.wall_time,
                    step=event.step,
                    device_name=device_name,
                    output_slot=output_slot,
                    node_name=match.group(1),
                    tensor_proto=value.tensor,
                    node_name_set=node_name_set,
                )
                if not health_pill:
                    continue
                mapping[health_pill.node_name].append(health_pill)

        # Keep reading events.
        return False

    def _process_health_pill_value(
        self,
        wall_time,
        step,
        device_name,
        output_slot,
        node_name,
        tensor_proto,
        node_name_set=None,
    ):
        """Creates a HealthPillEvent containing various properties of a health
        pill.

        Args:
          wall_time: The wall time in seconds.
          step: The session run step of the event.
          device_name: The name of the node's device.
          output_slot: The numeric output slot.
          node_name: The name of the node (without the output slot).
          tensor_proto: A tensor proto of data.
          node_name_set: An optional set of node names that are relevant. If not
            provided, no filtering by relevance occurs.

        Returns:
          An event_accumulator.HealthPillEvent. Or None if one could not be created.
        """
        if node_name_set and node_name not in node_name_set:
            # This event is not relevant.
            return None

        # Since we seek health pills for a specific step, this function
        # returns 1 health pill per node per step. The wall time is the
        # seconds since the epoch.
        elements = list(tensor_util.make_ndarray(tensor_proto))
        return HealthPillEvent(
            wall_time=wall_time,
            step=step,
            device_name=device_name,
            output_slot=output_slot,
            node_name=node_name,
            dtype=repr(tf.as_dtype(elements[12])),
            shape=elements[14:],
            value=elements,
        )

    @wrappers.Request.application
    def _serve_numerics_alert_report_handler(self, request):
        """A (wrapped) werkzeug handler for serving numerics alert report.

        Accepts GET requests and responds with an array of JSON-ified
        NumericsAlertReportRow.

        Each JSON-ified NumericsAlertReportRow object has the following format:
        {
            'device_name': string,
            'tensor_name': string,
            'first_timestamp': float,
            'nan_event_count': int,
            'neg_inf_event_count': int,
            'pos_inf_event_count': int
        }

        These objects are sorted by ascending order of first_timestamp in the
        response array.

        Args:
          request: The request, currently assumed to be empty.

        Returns:
          A werkzeug BaseResponse object.
        """
        if request.method != "GET":
            response = (
                "%s requests are forbidden by the debugger plugin."
                % request.method
            )
            return http_util.Respond(request, response, "text/plain", code=405)

        report = self._debugger_data_server.numerics_alert_report()

        # Convert the named tuples to dictionaries so we JSON them into objects.
        response = [
            r._asdict() for r in report
        ]  # pylint: disable=protected-access
        return http_util.Respond(request, response, "application/json")



# Copyright 2017 The TensorFlow Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""Receives data from a TensorFlow debugger. Writes event summaries.

This listener server writes debugging-related events into a logdir
directory, from which a TensorBoard instance can read.
"""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import collections
import functools
import json

from six.moves import queue

import tensorflow as tf
from tensorboard.plugins.debugger import comm_channel as comm_channel_lib
from tensorboard.plugins.debugger import debug_graphs_helper
from tensorboard.plugins.debugger import tensor_helper
from tensorboard.plugins.debugger import tensor_store as tensor_store_lib
from tensorboard.util import tb_logging
from tensorflow.core.debug import debug_service_pb2
from tensorflow.python import debug as tf_debug
from tensorflow.python.debug.lib import debug_data
from tensorflow.python.debug.lib import grpc_debug_server

logger = tb_logging.get_logger()

RunKey = collections.namedtuple(
    "RunKey", ["input_names", "output_names", "target_nodes"]
)


def _extract_device_name_from_event(event):
    """Extract device name from a tf.Event proto carrying tensor value."""
    plugin_data_content = json.loads(
        tf.compat.as_str(event.summary.value[0].metadata.plugin_data.content)
    )
    return plugin_data_content["device"]


def _comm_metadata(run_key, timestamp):
    return {
        "type": "meta",
        "timestamp": timestamp,
        "data": {"run_key": run_key,},
    }


UNINITIALIZED_TAG = "Uninitialized"
UNSUPPORTED_TAG = "Unsupported"
NA_TAG = "N/A"

STRING_ELEMENT_MAX_LEN = 40


def _comm_tensor_data(
    device_name,
    node_name,
    maybe_base_expanded_node_name,
    output_slot,
    debug_op,
    tensor_value,
    wall_time,
):
    """Create a dict() as the outgoing data in the tensor data comm route.

    Note: The tensor data in the comm route does not include the value of the
    tensor in its entirety in general. Only if a tensor satisfies the following
    conditions will its entire value be included in the return value of this
    method:
    1. Has a numeric data type (e.g., float32, int32) and has fewer than 5
       elements.
    2. Is a string tensor and has fewer than 5 elements. Each string element is
       up to 40 bytes.

    Args:
      device_name: Name of the device that the tensor is on.
      node_name: (Original) name of the node that produces the tensor.
      maybe_base_expanded_node_name: Possbily base-expanded node name.
      output_slot: Output slot number.
      debug_op: Name of the debug op.
      tensor_value: Value of the tensor, as a numpy.ndarray.
      wall_time: Wall timestamp for the tensor.

    Returns:
      A dict representing the tensor data.
    """
    output_slot = int(output_slot)
    logger.info(
        "Recording tensor value: %s, %d, %s", node_name, output_slot, debug_op
    )
    tensor_values = None
    if isinstance(tensor_value, debug_data.InconvertibleTensorProto):
        if not tensor_value.initialized:
            tensor_dtype = UNINITIALIZED_TAG
            tensor_shape = UNINITIALIZED_TAG
        else:
            tensor_dtype = UNSUPPORTED_TAG
            tensor_shape = UNSUPPORTED_TAG
        tensor_values = NA_TAG
    else:
        tensor_dtype = tensor_helper.translate_dtype(tensor_value.dtype)
        tensor_shape = tensor_value.shape

        # The /comm endpoint should respond with tensor values only if the tensor is
        # small enough. Otherwise, the detailed values sould be queried through a
        # dedicated tensor_data that supports slicing.
        if tensor_helper.numel(tensor_shape) < 5:
            _, _, tensor_values = tensor_helper.array_view(tensor_value)
            if tensor_dtype == "string" and tensor_value is not None:
                tensor_values = tensor_helper.process_buffers_for_display(
                    tensor_values, limit=STRING_ELEMENT_MAX_LEN
                )

    return {
        "type": "tensor",
        "timestamp": wall_time,
        "data": {
            "device_name": device_name,
            "node_name": node_name,
            "maybe_base_expanded_node_name": maybe_base_expanded_node_name,
            "output_slot": output_slot,
            "debug_op": debug_op,
            "dtype": tensor_dtype,
            "shape": tensor_shape,
            "values": tensor_values,
        },
    }


class RunStates(object):
    """A class that keeps track of state of debugged Session.run() calls."""

    def __init__(self, breakpoints_func=None):
        """Constructor of RunStates.

        Args:
          breakpoint_func: A callable of the signatuer:
            def breakpoint_func():
            which returns all the currently activated breakpoints.
        """
        # Maps from run key to debug_graphs_helper.DebugGraphWrapper instance.
        self._run_key_to_original_graphs = dict()
        self._run_key_to_debug_graphs = dict()

        if breakpoints_func:
            assert callable(breakpoints_func)
            self._breakpoints_func = breakpoints_func

    def add_graph(self, run_key, device_name, graph_def, debug=False):
        """Add a GraphDef.

        Args:
          run_key: A key for the run, containing information about the feeds,
            fetches, and targets.
          device_name: The name of the device that the `GraphDef` is for.
          graph_def: An instance of the `GraphDef` proto.
          debug: Whether `graph_def` consists of the debug ops.
        """
        graph_dict = (
            self._run_key_to_debug_graphs
            if debug
            else self._run_key_to_original_graphs
        )
        if not run_key in graph_dict:
            graph_dict[run_key] = dict()  # Mapping device_name to GraphDef.
        graph_dict[run_key][
            tf.compat.as_str(device_name)
        ] = debug_graphs_helper.DebugGraphWrapper(graph_def)

    def get_graphs(self, run_key, debug=False):
        """Get the runtime GraphDef protos associated with a run key.

        Args:
          run_key: A Session.run kay.
          debug: Whether the debugger-decoratedgraph is to be retrieved.

        Returns:
          A `dict` mapping device name to `GraphDef` protos.
        """
        graph_dict = (
            self._run_key_to_debug_graphs
            if debug
            else self._run_key_to_original_graphs
        )
        graph_wrappers = graph_dict.get(run_key, {})
        graph_defs = dict()
        for device_name, wrapper in graph_wrappers.items():
            graph_defs[device_name] = wrapper.graph_def
        return graph_defs

    def get_graph(self, run_key, device_name, debug=False):
        """Get the runtime GraphDef proto associated with a run key and a
        device.

        Args:
          run_key: A Session.run kay.
          device_name: Name of the device in question.
          debug: Whether the debugger-decoratedgraph is to be retrieved.

        Returns:
          A `GraphDef` proto.
        """
        return self.get_graphs(run_key, debug=debug).get(device_name, None)

    def get_breakpoints(self):
        """Obtain all the currently activated breakpoints."""
        return self._breakpoints_func()

    def get_gated_grpc_tensors(self, run_key, device_name):
        return self._run_key_to_debug_graphs[run_key][
            device_name
        ].get_gated_grpc_tensors()

    def get_maybe_base_expanded_node_name(
        self, node_name, run_key, device_name
    ):
        """Obtain possibly base-expanded node name.

        Base-expansion is the transformation of a node name which happens to be the
        name scope of other nodes in the same graph. For example, if two nodes,
        called 'a/b' and 'a/b/read' in a graph, the name of the first node will
        be base-expanded to 'a/b/(b)'.

        This method uses caching to avoid unnecessary recomputation.

        Args:
          node_name: Name of the node.
          run_key: The run key to which the node belongs.
          graph_def: GraphDef to which the node belongs.

        Raises:
          ValueError: If `run_key` and/or `device_name` do not exist in the record.
        """
        device_name = tf.compat.as_str(device_name)
        if run_key not in self._run_key_to_original_graphs:
            raise ValueError("Unknown run_key: %s" % run_key)
        if device_name not in self._run_key_to_original_graphs[run_key]:
            raise ValueError(
                'Unknown device for run key "%s": %s' % (run_key, device_name)
            )
        return self._run_key_to_original_graphs[run_key][
            device_name
        ].maybe_base_expanded_node_name(node_name)


class InteractiveDebuggerDataStreamHandler(
    grpc_debug_server.EventListenerBaseStreamHandler
):
    """Implementation of stream handler for debugger data.

    Each instance of this class is created by a InteractiveDebuggerDataServer
    upon a gRPC stream established between the debugged Session::Run() invocation
    in TensorFlow core runtime and the InteractiveDebuggerDataServer instance.

    Each instance of this class does the following:
      1) receives a core metadata Event proto during its constructor call.
      2) receives GraphDef Event proto(s) through its on_graph_def method.
      3) receives tensor value Event proto(s) through its on_value_event method.
    """

    def __init__(
        self, incoming_channel, outgoing_channel, run_states, tensor_store
    ):
        """Constructor of InteractiveDebuggerDataStreamHandler.

        Args:
          incoming_channel: An instance of FIFO queue, which manages incoming data,
            e.g., ACK signals from the client side unblock breakpoints.
          outgoing_channel: An instance of `CommChannel`, which manages outgoing
            data, i.e., data regarding the starting of Session.runs and hitting of
            tensor breakpoint.s
          run_states: An instance of `RunStates`, which keeps track of the states
            (graphs and breakpoints) of debugged Session.run() calls.
          tensor_store: An instance of `TensorStore`, which stores Tensor values
            from debugged Session.run() calls.
        """
        super(InteractiveDebuggerDataStreamHandler, self).__init__()

        self._incoming_channel = incoming_channel
        self._outgoing_channel = outgoing_channel
        self._run_states = run_states
        self._tensor_store = tensor_store

        self._run_key = None
        self._graph_defs = dict()  # A dict mapping device name to GraphDef.
        self._graph_defs_arrive_first = True

    def on_core_metadata_event(self, event):
        """Implementation of the core metadata-carrying Event proto callback.

        Args:
          event: An Event proto that contains core metadata about the debugged
            Session::Run() in its log_message.message field, as a JSON string.
            See the doc string of debug_data.DebugDumpDir.core_metadata for details.
        """
        core_metadata = json.loads(event.log_message.message)
        input_names = ",".join(core_metadata["input_names"])
        output_names = ",".join(core_metadata["output_names"])
        target_nodes = ",".join(core_metadata["target_nodes"])

        self._run_key = RunKey(input_names, output_names, target_nodes)
        if not self._graph_defs:
            self._graph_defs_arrive_first = False
        else:
            for device_name in self._graph_defs:
                self._add_graph_def(device_name, self._graph_defs[device_name])

        self._outgoing_channel.put(
            _comm_metadata(self._run_key, event.wall_time)
        )

        # Wait for acknowledgement from client. Blocks until an item is got.
        logger.info("on_core_metadata_event() waiting for client ack (meta)...")
        self._incoming_channel.get()
        logger.info("on_core_metadata_event() client ack received (meta).")

        # TODO(cais): If eager mode, this should return something to yield.

    def _add_graph_def(self, device_name, graph_def):
        self._run_states.add_graph(
            self._run_key,
            device_name,
            tf_debug.reconstruct_non_debug_graph_def(graph_def),
        )
        self._run_states.add_graph(
            self._run_key, device_name, graph_def, debug=True
        )

    def on_graph_def(self, graph_def, device_name, wall_time):
        """Implementation of the GraphDef-carrying Event proto callback.

        Args:
          graph_def: A GraphDef proto. N.B.: The GraphDef is from
            the core runtime of a debugged Session::Run() call, after graph
            partition. Therefore it may differ from the GraphDef available to
            the general TensorBoard. For example, the GraphDef in general
            TensorBoard may get partitioned for multiple devices (CPUs and GPUs),
            each of which will generate a GraphDef event proto sent to this
            method.
          device_name: Name of the device on which the graph was created.
          wall_time: An epoch timestamp (in microseconds) for the graph.
        """
        # For now, we do nothing with the graph def. However, we must define this
        # method to satisfy the handler's interface. Furthermore, we may use the
        # graph in the future (for instance to provide a graph if there is no graph
        # provided otherwise).
        del wall_time
        self._graph_defs[device_name] = graph_def

        if not self._graph_defs_arrive_first:
            self._add_graph_def(device_name, graph_def)
            self._incoming_channel.get()

    def on_value_event(self, event):
        """Records the summary values based on an updated message from the
        debugger.

        Logs an error message if writing the event to disk fails.

        Args:
          event: The Event proto to be processed.
        """
        if not event.summary.value:
            logger.info("The summary of the event lacks a value.")
            return None

        # The node name property in the event proto is actually a watch key, which
        # is a concatenation of several pieces of data.
        watch_key = event.summary.value[0].node_name
        tensor_value = debug_data.load_tensor_from_event(event)
        device_name = _extract_device_name_from_event(event)
        node_name, output_slot, debug_op = event.summary.value[
            0
        ].node_name.split(":")
        maybe_base_expanded_node_name = self._run_states.get_maybe_base_expanded_node_name(
            node_name, self._run_key, device_name
        )
        self._tensor_store.add(watch_key, tensor_value)
        self._outgoing_channel.put(
            _comm_tensor_data(
                device_name,
                node_name,
                maybe_base_expanded_node_name,
                output_slot,
                debug_op,
                tensor_value,
                event.wall_time,
            )
        )

        logger.info("on_value_event(): waiting for client ack (tensors)...")
        self._incoming_channel.get()
        logger.info("on_value_event(): client ack received (tensor).")

        # Determine if the particular debug watch key is in the current list of
        # breakpoints. If it is, send an EventReply() to unblock the debug op.
        if self._is_debug_node_in_breakpoints(event.summary.value[0].node_name):
            logger.info(
                "Sending empty EventReply for breakpoint: %s",
                event.summary.value[0].node_name,
            )
            # TODO(cais): Support receiving and sending tensor value from front-end.
            return debug_service_pb2.EventReply()
        return None

    def _is_debug_node_in_breakpoints(self, debug_node_key):
        node_name, output_slot, debug_op = debug_node_key.split(":")
        output_slot = int(output_slot)
        return (
            node_name,
            output_slot,
            debug_op,
        ) in self._run_states.get_breakpoints()


# TODO(cais): Consider moving to a seperate python module.
class SourceManager(object):
    """Manages source files and tracebacks involved in the debugged TF
    program."""

    def __init__(self):
        # A dict mapping file path to file content as a list of strings.
        self._source_file_content = dict()
        # A dict mapping file path to host name.
        self._source_file_host = dict()
        # A dict mapping file path to last modified timestamp.
        self._source_file_last_modified = dict()
        # A dict mapping file path to size in bytes.
        self._source_file_bytes = dict()
        # Keeps track f the traceback of the latest graph version.
        self._graph_traceback = None
        self._graph_version = -1

    def add_debugged_source_file(self, debugged_source_file):
        """Add a DebuggedSourceFile proto."""
        # TODO(cais): Should the key include a host name, for certain distributed
        #   cases?
        key = debugged_source_file.file_path
        self._source_file_host[key] = debugged_source_file.host
        self._source_file_last_modified[
            key
        ] = debugged_source_file.last_modified
        self._source_file_bytes[key] = debugged_source_file.bytes
        self._source_file_content[key] = debugged_source_file.lines

    def add_graph_traceback(self, graph_version, graph_traceback):
        if graph_version > self._graph_version:
            self._graph_traceback = graph_traceback
            self._graph_version = graph_version

    def get_paths(self):
        """Get the paths to all available source files."""
        return list(self._source_file_content.keys())

    def get_content(self, file_path):
        """Get the content of a source file.

        # TODO(cais): Maybe support getting a range of lines by line number.

        Args:
          file_path: Path to the source file.
        """
        return self._source_file_content[file_path]

    def get_op_traceback(self, op_name):
        """Get the traceback of an op in the latest version of the TF graph.

        Args:
          op_name: Name of the op.

        Returns:
          Creation traceback of the op, in the form of a list of 2-tuples:
            (file_path, lineno)

        Raises:
          ValueError: If the op with the given name cannot be found in the latest
            version of the graph that this SourceManager instance has received, or
            if this SourceManager instance has not received any graph traceback yet.
        """
        if not self._graph_traceback:
            raise ValueError("No graph traceback has been received yet.")
        for op_log_entry in self._graph_traceback.log_entries:
            if op_log_entry.name == op_name:
                return self._code_def_to_traceback_list(op_log_entry.code_def)
        raise ValueError(
            'No op named "%s" can be found in the graph of the latest version '
            " (%d)." % (op_name, self._graph_version)
        )

    def get_file_tracebacks(self, file_path):
        """Get the lists of ops created at lines of a specified source file.

        Args:
          file_path: Path to the source file.

        Returns:
          A dict mapping line number to a list of 2-tuples,
            `(op_name, stack_position)`
          `op_name` is the name of the name of the op whose creation traceback
            includes the line.
          `stack_position` is the position of the line in the op's creation
            traceback, represented as a 0-based integer.

        Raises:
          ValueError: If `file_path` does not point to a source file that has been
            received by this instance of `SourceManager`.
        """
        if file_path not in self._source_file_content:
            raise ValueError(
                'Source file of path "%s" has not been received by this instance of '
                "SourceManager." % file_path
            )

        lineno_to_op_names_and_stack_position = dict()
        for op_log_entry in self._graph_traceback.log_entries:
            for stack_pos, trace in enumerate(op_log_entry.code_def.traces):
                if (
                    self._graph_traceback.id_to_string[trace.file_id]
                    == file_path
                ):
                    if (
                        trace.lineno
                        not in lineno_to_op_names_and_stack_position
                    ):
                        lineno_to_op_names_and_stack_position[trace.lineno] = []
                    lineno_to_op_names_and_stack_position[trace.lineno].append(
                        (op_log_entry.name, stack_pos)
                    )
        return lineno_to_op_names_and_stack_position

    def _code_def_to_traceback_list(self, code_def):
        return [
            (self._graph_traceback.id_to_string[trace.file_id], trace.lineno)
            for trace in code_def.traces
        ]


class InteractiveDebuggerDataServer(
    grpc_debug_server.EventListenerBaseServicer
):
    """A service that receives and writes debugger data such as health
    pills."""

    def __init__(self, receive_port):
        """Receives health pills from a debugger and writes them to disk.

        Args:
          receive_port: The port at which to receive health pills from the
            TensorFlow debugger.
          always_flush: A boolean indicating whether the EventsWriter will be
            flushed after every write. Can be used for testing.
        """
        super(InteractiveDebuggerDataServer, self).__init__(
            receive_port, InteractiveDebuggerDataStreamHandler
        )

        self._incoming_channel = queue.Queue()
        self._outgoing_channel = comm_channel_lib.CommChannel()
        self._run_states = RunStates(breakpoints_func=lambda: self.breakpoints)
        self._tensor_store = tensor_store_lib.TensorStore()
        self._source_manager = SourceManager()

        curried_handler_constructor = functools.partial(
            InteractiveDebuggerDataStreamHandler,
            self._incoming_channel,
            self._outgoing_channel,
            self._run_states,
            self._tensor_store,
        )
        grpc_debug_server.EventListenerBaseServicer.__init__(
            self, receive_port, curried_handler_constructor
        )

    def SendTracebacks(self, request, context):
        self._source_manager.add_graph_traceback(
            request.graph_version, request.graph_traceback
        )
        return debug_service_pb2.EventReply()

    def SendSourceFiles(self, request, context):
        # TODO(cais): Handle case in which the size of the request is greater than
        #   the 4-MB gRPC limit.
        for source_file in request.source_files:
            self._source_manager.add_debugged_source_file(source_file)
        return debug_service_pb2.EventReply()

    def get_graphs(self, run_key, debug=False):
        return self._run_states.get_graphs(run_key, debug=debug)

    def get_graph(self, run_key, device_name, debug=False):
        return self._run_states.get_graph(run_key, device_name, debug=debug)

    def get_gated_grpc_tensors(self, run_key, device_name):
        return self._run_states.get_gated_grpc_tensors(run_key, device_name)

    def get_outgoing_message(self, pos):
        msg, _ = self._outgoing_channel.get(pos)
        return msg

    def put_incoming_message(self, message):
        return self._incoming_channel.put(message)

    def query_tensor_store(
        self, watch_key, time_indices=None, slicing=None, mapping=None
    ):
        """Query tensor store for a given debugged tensor value.

        Args:
          watch_key: The watch key of the debugged tensor being sought. Format:
            <node_name>:<output_slot>:<debug_op>
            E.g., Dense_1/MatMul:0:DebugIdentity.
          time_indices: Optional time indices string By default, the lastest time
            index ('-1') is returned.
          slicing: Optional slicing string.
          mapping: Optional mapping string, e.g., 'image/png'.

        Returns:
          If mapping is `None`, the possibly sliced values as a nested list of
            values or its mapped format. A `list` of nested `list` of values,
          If mapping is not `None`, the format of the return value will depend on
            the mapping.
        """
        return self._tensor_store.query(
            watch_key,
            time_indices=time_indices,
            slicing=slicing,
            mapping=mapping,
        )

    def query_source_file_paths(self):
        """Query the source files involved in the current debugged TF program.

        Returns:
          A `list` of file paths. The files that belong to the TensorFlow Python
            library itself are *not* included.
        """
        return self._source_manager.get_paths()

    def query_source_file_content(self, file_path):
        """Query the content of a given source file.

        # TODO(cais): Allow query only a range of the source lines.

        Returns:
          The source lines as a list of `str`.
        """
        return list(self._source_manager.get_content(file_path))

    def query_op_traceback(self, op_name):
        """Query the tracebacks of ops in a TensorFlow graph.

        Returns:
          TODO(cais):
        """
        return self._source_manager.get_op_traceback(op_name)

    def query_file_tracebacks(self, file_path):
        """Query the lists of ops created at lines of a given source file.

        Args:
          file_path: Path to the source file to get the tracebacks for.

        Returns:
          A `dict` mapping line number in the specified source file to a list of
            2-tuples:
              `(op_name, stack_position)`.
            `op_name` is the name of the name of the op whose creation traceback
              includes the line.
            `stack_position` is the position of the line in the op's creation
              traceback, represented as a 0-based integer.
        """
        return self._source_manager.get_file_tracebacks(file_path)

    def dispose(self):
        """Disposes of this object.

        Call only after this is done being used.
        """
        self._tensor_store.dispose()



# -*- coding: utf-8 -*-
# Copyright 2019 The TensorFlow Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""Utilities for handling Keras model in graph plugin.

Two canonical types of Keras model are Functional and Sequential.
A model can be serialized as JSON and deserialized to reconstruct a model.
This utility helps with dealing with the serialized Keras model.

They have distinct structures to the configurations in shapes below:
Functional:
  config
    name: Name of the model. If not specified, it is 'model' with
          an optional suffix if there are more than one instance.
    input_layers: Keras.layers.Inputs in the model.
    output_layers: Layer names that are outputs of the model.
    layers: list of layer configurations.
      layer: [*]
        inbound_nodes: inputs to this layer.

Sequential:
  config
    name: Name of the model. If not specified, it is 'sequential' with
          an optional suffix if there are more than one instance.
    layers: list of layer configurations.
      layer: [*]

[*]: Note that a model can be a layer.
Please refer to https://github.com/tensorflow/tfjs-layers/blob/master/src/keras_format/model_serialization.ts
for more complete definition.
"""
from tensorboard.compat.proto.graph_pb2 import GraphDef
from tensorboard.compat.tensorflow_stub import dtypes


def _walk_layers(keras_layer):
    """Walks the nested keras layer configuration in preorder.

    Args:
      keras_layer: Keras configuration from model.to_json.

    Yields:
      A tuple of (name_scope, layer_config).
      name_scope: a string representing a scope name, similar to that of tf.name_scope.
      layer_config: a dict representing a Keras layer configuration.
    """
    yield ("", keras_layer)
    if keras_layer.get("config").get("layers"):
        name_scope = keras_layer.get("config").get("name")
        for layer in keras_layer.get("config").get("layers"):
            for (sub_name_scope, sublayer) in _walk_layers(layer):
                sub_name_scope = (
                    "%s/%s" % (name_scope, sub_name_scope)
                    if sub_name_scope
                    else name_scope
                )
                yield (sub_name_scope, sublayer)


def _scoped_name(name_scope, node_name):
    """Returns scoped name for a node as a string in the form '<scope>/<node
    name>'.

    Args:
      name_scope: a string representing a scope name, similar to that of tf.name_scope.
      node_name: a string representing the current node name.

    Returns
      A string representing a scoped name.
    """
    if name_scope:
        return "%s/%s" % (name_scope, node_name)
    return node_name


def _is_model(layer):
    """Returns True if layer is a model.

    Args:
      layer: a dict representing a Keras model configuration.

    Returns:
      bool: True if layer is a model.
    """
    return layer.get("config").get("layers") is not None


def _norm_to_list_of_layers(maybe_layers):
    """Normalizes to a list of layers.

    Args:
      maybe_layers: A list of data[1] or a list of list of data.

    Returns:
      List of list of data.

    [1]: A Functional model has fields 'inbound_nodes' and 'output_layers' which can
    look like below:
    - ['in_layer_name', 0, 0]
    - [['in_layer_is_model', 1, 0], ['in_layer_is_model', 1, 1]]
    The data inside the list seems to describe [name, size, index].
    """
    return (
        maybe_layers if isinstance(maybe_layers[0], (list,)) else [maybe_layers]
    )


def _update_dicts(
    name_scope,
    model_layer,
    input_to_in_layer,
    model_name_to_output,
    prev_node_name,
):
    """Updates input_to_in_layer, model_name_to_output, and prev_node_name
    based on the model_layer.

    Args:
      name_scope: a string representing a scope name, similar to that of tf.name_scope.
      model_layer: a dict representing a Keras model configuration.
      input_to_in_layer: a dict mapping Keras.layers.Input to inbound layer.
      model_name_to_output: a dict mapping Keras Model name to output layer of the model.
      prev_node_name: a string representing a previous, in sequential model layout,
                      node name.

    Returns:
      A tuple of (input_to_in_layer, model_name_to_output, prev_node_name).
      input_to_in_layer: a dict mapping Keras.layers.Input to inbound layer.
      model_name_to_output: a dict mapping Keras Model name to output layer of the model.
      prev_node_name: a string representing a previous, in sequential model layout,
                      node name.
    """
    layer_config = model_layer.get("config")
    if not layer_config.get("layers"):
        raise ValueError("layer is not a model.")

    node_name = _scoped_name(name_scope, layer_config.get("name"))
    input_layers = layer_config.get("input_layers")
    output_layers = layer_config.get("output_layers")
    inbound_nodes = model_layer.get("inbound_nodes")

    is_functional_model = bool(input_layers and output_layers)
    # In case of [1] and the parent model is functional, current layer
    # will have the 'inbound_nodes' property.
    is_parent_functional_model = bool(inbound_nodes)

    if is_parent_functional_model and is_functional_model:
        for (input_layer, inbound_node) in zip(input_layers, inbound_nodes):
            input_layer_name = _scoped_name(node_name, input_layer)
            inbound_node_name = _scoped_name(name_scope, inbound_node[0])
            input_to_in_layer[input_layer_name] = inbound_node_name
    elif is_parent_functional_model and not is_functional_model:
        # Sequential model can take only one input. Make sure inbound to the
        # model is linked to the first layer in the Sequential model.
        prev_node_name = _scoped_name(name_scope, inbound_nodes[0][0][0])
    elif (
        not is_parent_functional_model
        and prev_node_name
        and is_functional_model
    ):
        assert len(input_layers) == 1, (
            "Cannot have multi-input Functional model when parent model "
            "is not Functional. Number of input layers: %d" % len(input_layer)
        )
        input_layer = input_layers[0]
        input_layer_name = _scoped_name(node_name, input_layer)
        input_to_in_layer[input_layer_name] = prev_node_name

    if is_functional_model and output_layers:
        layers = _norm_to_list_of_layers(output_layers)
        layer_names = [_scoped_name(node_name, layer[0]) for layer in layers]
        model_name_to_output[node_name] = layer_names
    else:
        last_layer = layer_config.get("layers")[-1]
        last_layer_name = last_layer.get("config").get("name")
        output_node = _scoped_name(node_name, last_layer_name)
        model_name_to_output[node_name] = [output_node]
    return (input_to_in_layer, model_name_to_output, prev_node_name)


def keras_model_to_graph_def(keras_layer):
    """Returns a GraphDef representation of the Keras model in a dict form.

    Note that it only supports models that implemented to_json().

    Args:
      keras_layer: A dict from Keras model.to_json().

    Returns:
      A GraphDef representation of the layers in the model.
    """
    input_to_layer = {}
    model_name_to_output = {}
    g = GraphDef()

    # Sequential model layers do not have a field "inbound_nodes" but
    # instead are defined implicitly via order of layers.
    prev_node_name = None

    for (name_scope, layer) in _walk_layers(keras_layer):
        if _is_model(layer):
            (
                input_to_layer,
                model_name_to_output,
                prev_node_name,
            ) = _update_dicts(
                name_scope,
                layer,
                input_to_layer,
                model_name_to_output,
                prev_node_name,
            )
            continue

        layer_config = layer.get("config")
        node_name = _scoped_name(name_scope, layer_config.get("name"))

        node_def = g.node.add()
        node_def.name = node_name

        if layer.get("class_name") is not None:
            keras_cls_name = layer.get("class_name").encode("ascii")
            node_def.attr["keras_class"].s = keras_cls_name

        if layer_config.get("dtype") is not None:
            tf_dtype = dtypes.as_dtype(layer_config.get("dtype"))
            node_def.attr["dtype"].type = tf_dtype.as_datatype_enum

        if layer.get("inbound_nodes") is not None:
            for maybe_inbound_node in layer.get("inbound_nodes"):
                inbound_nodes = _norm_to_list_of_layers(maybe_inbound_node)
                for [name, size, index, _] in inbound_nodes:
                    inbound_name = _scoped_name(name_scope, name)
                    # An input to a layer can be output from a model. In that case, the name
                    # of inbound_nodes to a layer is a name of a model. Remap the name of the
                    # model to output layer of the model. Also, since there can be multiple
                    # outputs in a model, make sure we pick the right output_layer from the model.
                    inbound_node_names = model_name_to_output.get(
                        inbound_name, [inbound_name]
                    )
                    node_def.input.append(inbound_node_names[index])
        elif prev_node_name is not None:
            node_def.input.append(prev_node_name)

        if node_name in input_to_layer:
            node_def.input.append(input_to_layer.get(node_name))

        prev_node_name = node_def.name

    return g



# Copyright 2017 The TensorFlow Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import numpy as np

import tensorflow as tf
from tensorboard.plugins.debugger import tensor_helper
from tensorboard.util import tb_logging
from tensorflow.python.debug.lib import debug_data

logger = tb_logging.get_logger()

# TODO(cais): Defer some tensor values to TensorBase, within a bytes limit,
#   before discarding them.


class _TensorValueDiscarded(object):
    def __init__(self, watch_key, time_index):
        self._watch_key = watch_key
        self._time_index = time_index

    @property
    def watch_key(self):
        return self._watch_key

    @property
    def time_index(self):
        return self._time_index

    @property
    def nbytes(self):
        return 0


class _WatchStore(object):
    """The store for a single debug tensor watch.

    Discards data according to pre-set byte limit.
    """

    def __init__(self, watch_key, mem_bytes_limit=10e6):
        """Constructor of _WatchStore.

        The overflowing works as follows:
        The most recent tensor values are stored in memory, up to `mem_bytes_limit`
        bytes. But at least one (the most recent) value is always stored in memory.
        For older tensors exceeding that limit, they are discarded.

        Args:
          watch_key: A string representing the debugger tensor watch, with th
            format:
              <NODE_NAME>:<OUTPUT_SLOT>:<DEBUG_OP_NAME>
            e.g.,
              'Dense_1/BiasAdd:0:DebugIdentity'.
          mem_bytes_limit: Limit on number of bytes to store in memory.
        """

        self._watch_key = watch_key
        self._mem_bytes_limit = mem_bytes_limit
        self._in_mem_bytes = 0
        self._disposed = False
        self._data = []  # A map from index to tensor value.

    def add(self, value):
        """Add a tensor the watch store."""
        if self._disposed:
            raise ValueError(
                "Cannot add value: this _WatchStore instance is already disposed"
            )
        self._data.append(value)
        if hasattr(value, "nbytes"):
            self._in_mem_bytes += value.nbytes
            self._ensure_bytes_limits()

    def _ensure_bytes_limits(self):
        # TODO(cais): Thread safety?
        if self._in_mem_bytes <= self._mem_bytes_limit:
            return

        i = len(self._data) - 1
        cum_mem_size = 0
        while i >= 0:
            if hasattr(self._data[i], "nbytes"):
                cum_mem_size += self._data[i].nbytes
            if i < len(self._data) - 1 and cum_mem_size > self._mem_bytes_limit:
                # Always keep at least one time index in the memory.
                break
            i -= 1
        # i is now the last time index to discard.

        # Mark remaining ones as discarded.
        while i >= 0:
            if not isinstance(self._data[i], _TensorValueDiscarded):
                self._data[i] = _TensorValueDiscarded(self._watch_key, i)
            i -= 1

    def num_total(self):
        """Get the total number of values."""
        return len(self._data)

    def num_in_memory(self):
        """Get number of values in memory."""
        n = len(self._data) - 1
        while n >= 0:
            if isinstance(self._data[n], _TensorValueDiscarded):
                break
            n -= 1
        return len(self._data) - 1 - n

    def num_discarded(self):
        """Get the number of values discarded due to exceeding both limits."""
        if not self._data:
            return 0
        n = 0
        while n < len(self._data):
            if not isinstance(self._data[n], _TensorValueDiscarded):
                break
            n += 1
        return n

    def query(self, time_indices):
        """Query the values at given time indices.

        Args:
          time_indices: 0-based time indices to query, as a `list` of `int`.

        Returns:
          Values as a list of `numpy.ndarray` (for time indices in memory) or
          `None` (for time indices discarded).
        """
        if self._disposed:
            raise ValueError(
                "Cannot query: this _WatchStore instance is already disposed"
            )
        if not isinstance(time_indices, (tuple, list)):
            time_indices = [time_indices]
        output = []
        for time_index in time_indices:
            if isinstance(self._data[time_index], _TensorValueDiscarded):
                output.append(None)
            else:
                data_item = self._data[time_index]
                if (
                    hasattr(data_item, "dtype")
                    and tensor_helper.translate_dtype(data_item.dtype)
                    == "string"
                ):
                    _, _, data_item = tensor_helper.array_view(data_item)
                    data_item = np.array(
                        tensor_helper.process_buffers_for_display(data_item),
                        dtype=np.object,
                    )
                output.append(data_item)

        return output

    def dispose(self):
        self._disposed = True


class TensorStore(object):
    def __init__(self, watch_mem_bytes_limit=10e6):
        """Constructor of TensorStore.

        Args:
          watch_mem_bytes_limit: Limit on number of bytes to store in memory for
            each watch key.
        """
        self._watch_mem_bytes_limit = watch_mem_bytes_limit
        self._tensor_data = (
            dict()
        )  # A map from watch key to _WatchStore instances.

    def add(self, watch_key, tensor_value):
        """Add a tensor value.

        Args:
          watch_key: A string representing the debugger tensor watch, e.g.,
            'Dense_1/BiasAdd:0:DebugIdentity'.
          tensor_value: The value of the tensor as a numpy.ndarray.
        """
        if watch_key not in self._tensor_data:
            self._tensor_data[watch_key] = _WatchStore(
                watch_key, mem_bytes_limit=self._watch_mem_bytes_limit
            )
        self._tensor_data[watch_key].add(tensor_value)

    def query(self, watch_key, time_indices=None, slicing=None, mapping=None):
        """Query tensor store for a given watch_key.

        Args:
          watch_key: The watch key to query.
          time_indices: A numpy-style slicing string for time indices. E.g.,
            `-1`, `:-2`, `[::2]`. If not provided (`None`), will use -1.
          slicing: A numpy-style slicing string for individual time steps.
          mapping: An mapping string or a list of them. Supported mappings:
            `{None, 'image/png', 'health-pill'}`.

        Returns:
          The potentially sliced values as a nested list of values or its mapped
            format. A `list` of nested `list` of values.

        Raises:
          ValueError: If the shape of the sliced array is incompatible with mapping
            mode. Or if the mapping type is invalid.
        """
        if watch_key not in self._tensor_data:
            raise KeyError("watch_key not found: %s" % watch_key)

        if time_indices is None:
            time_indices = "-1"
        time_slicing = tensor_helper.parse_time_indices(time_indices)
        all_time_indices = list(range(self._tensor_data[watch_key].num_total()))
        sliced_time_indices = all_time_indices[time_slicing]
        if not isinstance(sliced_time_indices, list):
            sliced_time_indices = [sliced_time_indices]

        recombine_and_map = False
        step_mapping = mapping
        if len(sliced_time_indices) > 1 and mapping not in (None,):
            recombine_and_map = True
            step_mapping = None

        output = []
        for index in sliced_time_indices:
            value = self._tensor_data[watch_key].query(index)[0]
            if value is not None and not isinstance(
                value, debug_data.InconvertibleTensorProto
            ):
                output.append(
                    tensor_helper.array_view(
                        value, slicing=slicing, mapping=step_mapping
                    )[2]
                )
            else:
                output.append(None)

        if recombine_and_map:
            if mapping == "image/png":
                output = tensor_helper.array_to_base64_png(output)
            elif mapping and mapping != "none":
                logger.warning(
                    "Unsupported mapping mode after recomining time steps: %s",
                    mapping,
                )
        return output

    def dispose(self):
        for watch_key in self._tensor_data:
            self._tensor_data[watch_key].dispose()



# Copyright 2020 The TensorFlow Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""Utilities for measuring elapsed time."""

import contextlib
import logging
import threading
import time

from tensorboard.util import tb_logging

logger = tb_logging.get_logger()


def log_latency(region_name_or_function_to_decorate, log_level=None):
    """Log latency in a function or region.

    Three usages are supported. As a decorator:

    >>> @log_latency
    ... def function_1():
    ...     pass
    ...


    As a decorator with a custom label for the region:

    >>> @log_latency("custom_label")
    ... def function_2():
    ...     pass
    ...

    As a context manager:

    >>> def function_3():
    ...     with log_latency("region_within_function"):
    ...         pass
    ...

    Args:
        region_name_or_function_to_decorate: Either: a `str`, in which
            case the result of this function may be used as either a
            decorator or a context manager; or a callable, in which case
            the result of this function is a decorated version of that
            callable.
        log_level: Optional integer logging level constant. Defaults to
            `logging.INFO`.

    Returns:
        A decorated version of the input callable, or a dual
        decorator/context manager with the input region name.
    """

    if log_level is None:
        log_level = logging.INFO

    if isinstance(region_name_or_function_to_decorate, str):
        region_name = region_name_or_function_to_decorate
        return _log_latency(region_name, log_level)
    else:
        function_to_decorate = region_name_or_function_to_decorate
        qualname = getattr(function_to_decorate, "__qualname__", None)
        if qualname is None:
            qualname = str(function_to_decorate)
        decorator = _log_latency(qualname, log_level)
        return decorator(function_to_decorate)


class _ThreadLocalStore(threading.local):
    def __init__(self):
        self.nesting_level = 0


_store = _ThreadLocalStore()


@contextlib.contextmanager
def _log_latency(name, log_level):
    if not logger.isEnabledFor(log_level):
        yield
        return

    start_level = _store.nesting_level
    try:
        started = time.time()
        _store.nesting_level = start_level + 1
        indent = (" " * 2) * start_level
        thread = threading.current_thread()
        prefix = "%s[%x]%s" % (thread.name, thread.ident, indent)
        _log(log_level, "%s ENTER %s", prefix, name)
        yield
    finally:
        _store.nesting_level = start_level
        elapsed = time.time() - started
        _log(
            log_level, "%s LEAVE %s - %0.6fs elapsed", prefix, name, elapsed,
        )


def _log(log_level, msg, *args):
    # Forwarding method to ensure that all logging statements
    # originating in this module have the same line number; if the
    # "ENTER" log is on a line with 2-digit number and the "LEAVE" log
    # is on a line with 3-digit number, the logs are misaligned and
    # harder to read.
    logger.log(log_level, msg, *args)



# Copyright 2019 The TensorFlow Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================

"""Implementation for a multi-file directory loader."""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

from tensorboard.backend.event_processing import directory_watcher
from tensorboard.backend.event_processing import io_wrapper
from tensorboard.compat import tf
from tensorboard.util import tb_logging


logger = tb_logging.get_logger()


# Sentinel object for an inactive path.
_INACTIVE = object()


class DirectoryLoader(object):
    """Loader for an entire directory, maintaining multiple active file
    loaders.

    This class takes a directory, a factory for loaders, and optionally a
    path filter and watches all the paths inside that directory for new data.
    Each file loader created by the factory must read a path and produce an
    iterator of (timestamp, value) pairs.

    Unlike DirectoryWatcher, this class does not assume that only one file
    receives new data at a time; there can be arbitrarily many active files.
    However, any file whose maximum load timestamp fails an "active" predicate
    will be marked as inactive and no longer checked for new data.
    """

    def __init__(
        self,
        directory,
        loader_factory,
        path_filter=lambda x: True,
        active_filter=lambda timestamp: True,
    ):
        """Constructs a new MultiFileDirectoryLoader.

        Args:
          directory: The directory to load files from.
          loader_factory: A factory for creating loaders. The factory should take a
            path and return an object that has a Load method returning an iterator
            yielding (unix timestamp as float, value) pairs for any new data
          path_filter: If specified, only paths matching this filter are loaded.
          active_filter: If specified, any loader whose maximum load timestamp does
            not pass this filter will be marked as inactive and no longer read.

        Raises:
          ValueError: If directory or loader_factory are None.
        """
        if directory is None:
            raise ValueError("A directory is required")
        if loader_factory is None:
            raise ValueError("A loader factory is required")
        self._directory = directory
        self._loader_factory = loader_factory
        self._path_filter = path_filter
        self._active_filter = active_filter
        self._loaders = {}
        self._max_timestamps = {}

    def Load(self):
        """Loads new values from all active files.

        Yields:
          All values that have not been yielded yet.

        Raises:
          DirectoryDeletedError: If the directory has been permanently deleted
            (as opposed to being temporarily unavailable).
        """
        try:
            all_paths = io_wrapper.ListDirectoryAbsolute(self._directory)
            paths = sorted(p for p in all_paths if self._path_filter(p))
            for path in paths:
                for value in self._LoadPath(path):
                    yield value
        except tf.errors.OpError as e:
            if not tf.io.gfile.exists(self._directory):
                raise directory_watcher.DirectoryDeletedError(
                    "Directory %s has been permanently deleted"
                    % self._directory
                )
            else:
                logger.info("Ignoring error during file loading: %s" % e)

    def _LoadPath(self, path):
        """Generator for values from a single path's loader.

        Args:
          path: the path to load from

        Yields:
          All values from this path's loader that have not been yielded yet.
        """
        max_timestamp = self._max_timestamps.get(path, None)
        if max_timestamp is _INACTIVE or self._MarkIfInactive(
            path, max_timestamp
        ):
            logger.debug("Skipping inactive path %s", path)
            return
        loader = self._loaders.get(path, None)
        if loader is None:
            try:
                loader = self._loader_factory(path)
            except tf.errors.NotFoundError:
                # Happens if a file was removed after we listed the directory.
                logger.debug("Skipping nonexistent path %s", path)
                return
            self._loaders[path] = loader
        logger.info("Loading data from path %s", path)
        for timestamp, value in loader.Load():
            if max_timestamp is None or timestamp > max_timestamp:
                max_timestamp = timestamp
            yield value
        if not self._MarkIfInactive(path, max_timestamp):
            self._max_timestamps[path] = max_timestamp

    def _MarkIfInactive(self, path, max_timestamp):
        """If max_timestamp is inactive, returns True and marks the path as
        such."""
        logger.debug("Checking active status of %s at %s", path, max_timestamp)
        if max_timestamp is not None and not self._active_filter(max_timestamp):
            self._max_timestamps[path] = _INACTIVE
            del self._loaders[path]
            return True
        return False



# Copyright 2017 The TensorFlow Authors. All Rights Reserved.
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
# ==============================================================================
"""The TensorBoard Graphs plugin."""

from __future__ import absolute_import
from __future__ import division
from __future__ import print_function

import json
import six
from werkzeug import wrappers

from tensorboard import plugin_util
from tensorboard import context
from tensorboard.backend import http_util
from tensorboard.backend import process_graph
from tensorboard.backend.event_processing import tag_types
from tensorboard.compat.proto import config_pb2
from tensorboard.compat.proto import graph_pb2
from tensorboard.data import provider
from tensorboard.plugins import base_plugin
from tensorboard.plugins.graph import graph_util
from tensorboard.plugins.graph import keras_util
from tensorboard.plugins.graph import metadata
from tensorboard.util import tb_logging

logger = tb_logging.get_logger()

# The Summary API is implemented in TensorFlow because it uses TensorFlow internal APIs.
# As a result, this SummaryMetadata is a bit unconventional and uses non-public
# hardcoded name as the plugin name. Please refer to link below for the summary ops.
# https://github.com/tensorflow/tensorflow/blob/11f4ecb54708865ec757ca64e4805957b05d7570/tensorflow/python/ops/summary_ops_v2.py#L757
_PLUGIN_NAME_RUN_METADATA = "graph_run_metadata"
# https://github.com/tensorflow/tensorflow/blob/11f4ecb54708865ec757ca64e4805957b05d7570/tensorflow/python/ops/summary_ops_v2.py#L788
_PLUGIN_NAME_RUN_METADATA_WITH_GRAPH = "graph_run_metadata_graph"
# https://github.com/tensorflow/tensorflow/blob/565952cc2f17fdfd995e25171cf07be0f6f06180/tensorflow/python/ops/summary_ops_v2.py#L825
_PLUGIN_NAME_KERAS_MODEL = "graph_keras_model"


class GraphsPlugin(base_plugin.TBPlugin):
    """Graphs Plugin for TensorBoard."""

    plugin_name = metadata.PLUGIN_NAME

    def __init__(self, context):
        """Instantiates GraphsPlugin via TensorBoard core.

        Args:
          context: A base_plugin.TBContext instance.
        """
        self._multiplexer = context.multiplexer
        if context.flags and context.flags.generic_data == "true":
            self._data_provider = context.data_provider
        else:
            self._data_provider = None

    def get_plugin_apps(self):
        return {
            "/graph": self.graph_route,
            "/info": self.info_route,
            "/run_metadata": self.run_metadata_route,
        }

    def is_active(self):
        """The graphs plugin is active iff any run has a graph or metadata."""
        if self._data_provider:
            return False  # `list_plugins` as called by TB core suffices

        empty_context = context.RequestContext()  # not used
        return bool(self.info_impl(empty_context))

    def frontend_metadata(self):
        return base_plugin.FrontendMetadata(
            element_name="tf-graph-dashboard",
            # TODO(@chihuahua): Reconcile this setting with Health Pills.
            disable_reload=True,
        )

    def info_impl(self, ctx, experiment=None):
        """Returns a dict of all runs and their data availabilities."""
        result = {}

        def add_row_item(run, tag=None):
            run_item = result.setdefault(
                run,
                {
                    "run": run,
                    "tags": {},
                    # A run-wide GraphDef of ops.
                    "run_graph": False,
                },
            )

            tag_item = None
            if tag:
                tag_item = run_item.get("tags").setdefault(
                    tag,
                    {
                        "tag": tag,
                        "conceptual_graph": False,
                        # A tagged GraphDef of ops.
                        "op_graph": False,
                        "profile": False,
                    },
                )
            return (run_item, tag_item)

        if self._data_provider:
            mapping = self._data_provider.list_blob_sequences(
                ctx, experiment_id=experiment, plugin_name=metadata.PLUGIN_NAME,
            )
            for (run_name, tag_to_time_series) in six.iteritems(mapping):
                for tag in tag_to_time_series:
                    if tag == metadata.RUN_GRAPH_NAME:
                        (run_item, _) = add_row_item(run_name, None)
                        run_item["run_graph"] = True
                    else:
                        (_, tag_item) = add_row_item(run_name, tag)
                        tag_item["op_graph"] = True
            return result

        mapping = self._multiplexer.PluginRunToTagToContent(
            _PLUGIN_NAME_RUN_METADATA_WITH_GRAPH
        )
        for run_name, tag_to_content in six.iteritems(mapping):
            for (tag, content) in six.iteritems(tag_to_content):
                # The Summary op is defined in TensorFlow and does not use a stringified proto
                # as a content of plugin data. It contains single string that denotes a version.
                # https://github.com/tensorflow/tensorflow/blob/11f4ecb54708865ec757ca64e4805957b05d7570/tensorflow/python/ops/summary_ops_v2.py#L789-L790
                if content != b"1":
                    logger.warning(
                        "Ignoring unrecognizable version of RunMetadata."
                    )
                    continue
                (_, tag_item) = add_row_item(run_name, tag)
                tag_item["op_graph"] = True

        # Tensors associated with plugin name _PLUGIN_NAME_RUN_METADATA contain
        # both op graph and profile information.
        mapping = self._multiplexer.PluginRunToTagToContent(
            _PLUGIN_NAME_RUN_METADATA
        )
        for run_name, tag_to_content in six.iteritems(mapping):
            for (tag, content) in six.iteritems(tag_to_content):
                if content != b"1":
                    logger.warning(
                        "Ignoring unrecognizable version of RunMetadata."
                    )
                    continue
                (_, tag_item) = add_row_item(run_name, tag)
                tag_item["profile"] = True
                tag_item["op_graph"] = True

        # Tensors associated with plugin name _PLUGIN_NAME_KERAS_MODEL contain
        # serialized Keras model in JSON format.
        mapping = self._multiplexer.PluginRunToTagToContent(
            _PLUGIN_NAME_KERAS_MODEL
        )
        for run_name, tag_to_content in six.iteritems(mapping):
            for (tag, content) in six.iteritems(tag_to_content):
                if content != b"1":
                    logger.warning(
                        "Ignoring unrecognizable version of RunMetadata."
                    )
                    continue
                (_, tag_item) = add_row_item(run_name, tag)
                tag_item["conceptual_graph"] = True

        for (run_name, run_data) in six.iteritems(self._multiplexer.Runs()):
            if run_data.get(tag_types.GRAPH):
                (run_item, _) = add_row_item(run_name, None)
                run_item["run_graph"] = True

        for (run_name, run_data) in six.iteritems(self._multiplexer.Runs()):
            if tag_types.RUN_METADATA in run_data:
                for tag in run_data[tag_types.RUN_METADATA]:
                    (_, tag_item) = add_row_item(run_name, tag)
                    tag_item["profile"] = True

        return result

    def graph_impl(
        self,
        ctx,
        run,
        tag,
        is_conceptual,
        experiment=None,
        limit_attr_size=None,
        large_attrs_key=None,
    ):
        """Result of the form `(body, mime_type)`, or `None` if no graph
        exists."""
        if self._data_provider:
            if tag is None:
                tag = metadata.RUN_GRAPH_NAME
            graph_blob_sequences = self._data_provider.read_blob_sequences(
                ctx,
                experiment_id=experiment,
                plugin_name=metadata.PLUGIN_NAME,
                run_tag_filter=provider.RunTagFilter(runs=[run], tags=[tag]),
                downsample=1,
            )
            blob_datum_list = graph_blob_sequences.get(run, {}).get(tag, ())
            try:
                blob_ref = blob_datum_list[0].values[0]
            except IndexError:
                return None
            # Always use the blob_key approach for now, even if there is a direct url.
            graph_raw = self._data_provider.read_blob(
                ctx, blob_key=blob_ref.blob_key
            )
            # This method ultimately returns pbtxt, but we have to deserialize and
            # later reserialize this anyway, because a) this way we accept binary
            # protobufs too, and b) below we run `prepare_graph_for_ui` on the graph.
            graph = graph_pb2.GraphDef.FromString(graph_raw)

        elif is_conceptual:
            tensor_events = self._multiplexer.Tensors(run, tag)
            # Take the first event if there are multiple events written from different
            # steps.
            keras_model_config = json.loads(
                tensor_events[0].tensor_proto.string_val[0]
            )
            graph = keras_util.keras_model_to_graph_def(keras_model_config)

        elif tag:
            tensor_events = self._multiplexer.Tensors(run, tag)
            # Take the first event if there are multiple events written from different
            # steps.
            run_metadata = config_pb2.RunMetadata.FromString(
                tensor_events[0].tensor_proto.string_val[0]
            )
            graph = graph_pb2.GraphDef()

            for func_graph in run_metadata.function_graphs:
                graph_util.combine_graph_defs(
                    graph, func_graph.pre_optimization_graph
                )
        else:
            graph = self._multiplexer.Graph(run)

        # This next line might raise a ValueError if the limit parameters
        # are invalid (size is negative, size present but key absent, etc.).
        process_graph.prepare_graph_for_ui(
            graph, limit_attr_size, large_attrs_key
        )
        return (str(graph), "text/x-protobuf")  # pbtxt

    def run_metadata_impl(self, run, tag):
        """Result of the form `(body, mime_type)`, or `None` if no data
        exists."""
        if self._data_provider:
            # TODO(davidsoergel, wchargin): Consider plumbing run metadata through data providers.
            return None
        try:
            run_metadata = self._multiplexer.RunMetadata(run, tag)
        except ValueError:
            # TODO(stephanwlee): Should include whether FE is fetching for v1 or v2 RunMetadata
            # so we can remove this try/except.
            tensor_events = self._multiplexer.Tensors(run, tag)
            if tensor_events is None:
                return None
            # Take the first event if there are multiple events written from different
            # steps.
            run_metadata = config_pb2.RunMetadata.FromString(
                tensor_events[0].tensor_proto.string_val[0]
            )
        if run_metadata is None:
            return None
        return (str(run_metadata), "text/x-protobuf")  # pbtxt

    @wrappers.Request.application
    def info_route(self, request):
        ctx = plugin_util.context(request.environ)
        experiment = plugin_util.experiment_id(request.environ)
        info = self.info_impl(ctx, experiment)
        return http_util.Respond(request, info, "application/json")

    @wrappers.Request.application
    def graph_route(self, request):
        """Given a single run, return the graph definition in protobuf
        format."""
        ctx = plugin_util.context(request.environ)
        experiment = plugin_util.experiment_id(request.environ)
        run = request.args.get("run")
        tag = request.args.get("tag")
        conceptual_arg = request.args.get("conceptual", False)
        is_conceptual = True if conceptual_arg == "true" else False

        if run is None:
            return http_util.Respond(
                request, 'query parameter "run" is required', "text/plain", 400
            )

        limit_attr_size = request.args.get("limit_attr_size", None)
        if limit_attr_size is not None:
            try:
                limit_attr_size = int(limit_attr_size)
            except ValueError:
                return http_util.Respond(
                    request,
                    "query parameter `limit_attr_size` must be an integer",
                    "text/plain",
                    400,
                )

        large_attrs_key = request.args.get("large_attrs_key", None)

        try:
            result = self.graph_impl(
                ctx,
                run,
                tag,
                is_conceptual,
                experiment,
                limit_attr_size,
                large_attrs_key,
            )
        except ValueError as e:
            return http_util.Respond(request, e.message, "text/plain", code=400)
        else:
            if result is not None:
                (
                    body,
                    mime_type,
                ) = result  # pylint: disable=unpacking-non-sequence
                return http_util.Respond(request, body, mime_type)
            else:
                return http_util.Respond(
                    request, "404 Not Found", "text/plain", code=404
                )

    @wrappers.Request.application
    def run_metadata_route(self, request):
        """Given a tag and a run, return the session.run() metadata."""
        tag = request.args.get("tag")
        run = request.args.get("run")
        if tag is None:
            return http_util.Respond(
                request, 'query parameter "tag" is required', "text/plain", 400
            )
        if run is None:
            return http_util.Respond(
                request, 'query parameter "run" is required', "text/plain", 400
            )
        result = self.run_metadata_impl(run, tag)
        if result is not None:
            (body, mime_type) = result  # pylint: disable=unpacking-non-sequence
            return http_util.Respond(request, body, mime_type)
        else:
            return http_util.Respond(
                request, "404 Not Found", "text/plain", code=404
            )
